#
#     Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#     Lead Developers: Jay Baxter and Dan Lovell
#     Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#     Research Leads: Vikash Mansinghka, Patrick Shafto
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

import time
import inspect
import os
import json
import datetime
import re
import operator
import copy
import math
import ast
import sys
import random
import threading
import multiprocessing
import Queue
import pandas

import pylab
import numpy
import matplotlib.cm
from collections import defaultdict

import bayesdb.settings as S
from persistence_layer import PersistenceLayer
import api_utils
import data_utils
import utils
import pairwise
import functions as funcs
import select_utils
import estimate_columns_utils
import plotting_utils
import parser as p

import bayesdb.diagnostics_utils as diag_utils

class Engine(object):
    def __init__(self, crosscat_host=None, crosscat_port=8007, crosscat_engine_type='local', seed=None, testing=False, **kwargs):
        """ One optional argument that you may find yourself using frequently is seed.
        It defaults to random seed, but for testing/reproduceability purposes you may
        want a deterministic one. """

        self.persistence_layer = PersistenceLayer()
        self.parser = p.Parser()
        self.analyze_threads = dict() # Maps tablenames to whether they are doing an ANALYZE.
        self.testing = testing

        if crosscat_host is None or crosscat_host == 'localhost':
            self.online = False

            # Only dependent on CrossCat when you actually instantiate Engine
            # (i.e., allow engine to be imported in order to examine the API, without CrossCat)
            from crosscat.CrossCatClient import get_CrossCatClient
            self.backend = get_CrossCatClient(crosscat_engine_type, seed=seed, **kwargs)
        else:
            self.online = True
            self.hostname = crosscat_host
            self.port = crosscat_port
            self.URI = 'http://' + self.hostname + ':%d' % self.port

    def call_backend(self, method_name, args_dict):
        """
        Helper function used to call the CrossCat backend, whether it is remote or local.
        Accept method name and arguments for that method as input.
        """
        if self.online:
            out, id = api_utils.call(method_name, args_dict, self.URI)
        else:
            method = getattr(self.backend, method_name)
            out = method(**args_dict)
        return out

    def help(self, method=None):
        help_methods = dict()
        help_methods['create'] = """
        CREATE BTABLE <btable> FROM <filename.csv> [WITH CODEBOOK <codebook.csv>]

        CREATE COLUMN LIST <col1>[, <col2>...] FROM <btable> AS <column_list>
        """

        help_methods['select']="""
        SELECT <columns|functions> FROM <btable> [WHERE <whereclause>] [ORDER BY <columns|functions>] [LIMIT <limit>]
        """
        help_methods['infer'] = """
        INFER <columns|functions> FROM <btable> [WHERE <whereclause>] [WITH CONFIDENCE <confidence>] [WITH <numsamples> SAMPLES] [ORDER BY <columns|functions>] [LIMIT <limit>]
        """
        help_methods['simulate'] = """
        SIMULATE <columns> FROM <btable> [WHERE <whereclause>] [GIVEN <column>=<value>] TIMES <times> [SAVE TO <file>]
        """
        help_methods['estimate'] = """
        ESTIMATE COLUMNS FROM <btable> [WHERE <whereclause>] [ORDER BY <functions>] [LIMIT <limit>] [AS <column_list>]

        ESTIMATE PAIRWISE <function> FROM <btable> [FOR <columns>] [SAVE TO <file>] [SAVE CONNECTED COMPONENTS WITH THRESHOLD <threshold> AS <column_list>]

        ESTIMATE PAIRWISE ROW SIMILARITY [WITH RESPECT TO <columns|column_lists>]FROM <btable> [FOR <rows>] [SAVE TO <file>] [SAVE CONNECTED COMPONENTS WITH THRESHOLD <threshold> [INTO|AS] <btable>]
        """

        help_methods['execute'] = """
        EXECUTE FILE <filename.bql>
        """

        help_methods['update'] = """
        UPDATE SCHEMA FOR <btable> SET <col1>=<type1>[,<col2>=<type2>...]
        """

        help_methods['initialize'] = """
        INITIALIZE <num_models> MODELS FOR <btable> [WITH CONFIG <config>]
        """

        help_methods['analyze'] = """
        ANALYZE <btable> [MODEL[S] <model_index>-<model_index>] FOR (<num_iterations> ITERATIONS | <seconds> SECONDS)
        """

        help_methods['list'] = """
        LIST BTABLES
        """

        help_methods['show'] = """
        SHOW SCHEMA FOR <btable>

        SHOW MODELS FOR <btable>

        SHOW DIAGNOSTICS FOR <btable>

        SHOW COLUMN LISTS FOR <btable>

        SHOW COLUMNS <column_list> FOR <btable>

        SHOW ROW LISTS FOR <table>

        SHOW METADATA FOR <btable> [<metadata-key1> [, <metadata-key2>...]]

        SHOW LABEL FOR <btable> [<column-name-1> [, <column-name-2>...]]
        """

        help_methods['load'] = """
        LOAD MODELS <filename.pkl.gz> INTO <btable>
        """

        help_methods['save'] = """
        SAVE MODELS FROM <btable> TO <filename.pkl.gz>
        """

        help_methods['drop'] = """
        DROP BTABLE <btable>

        DROP MODEL[S] [<model_index>-<model_index>] FROM <btable>
        """

        help_methods['summarize'] = """
        SUMMARIZE <SELECT|INFER|SIMULATE>
        """

        help_methods['plot'] = """
        PLOT <SELECT|INFER|SIMULATE>
        """

        ## Important to make sure that http://probcomp.csail.mit.edu/bayesdb/docs/bql.html is up to date
        help_all = """
        Welcome to BQL help.

        For the BQL documentation, please visit:
        http://probcomp.csail.mit.edu/bayesdb/docs/bql.html

        Here is a list of BQL commands and their syntax:
        """
        help_basic = """
        Welcome to BQL help.

        If you know the query you wish to use, type 'HELP <query_name>'

        For all queries, type 'HELP ALL'

        For the complete BQL documentation, please visit:
        http://probcomp.csail.mit.edu/bayesdb/docs/bql.html
        """

        if method == None:
            help_string = help_basic
        elif method == 'all':
            help_string = help_all + ''.join(help_methods.values())
        elif method in help_methods:
            help_string = help_all + help_methods[method]
        else:
            help_string = help_basic + '''
        The method you typed was not recognize. Try "HELP", "HELP ALL",
        or one of the following:
        \tHELP ''' + '\n\tHELP '.join(help_methods.keys()).upper()

        return dict(message=help_string)

    def is_analyzing(self, tablename):
        return (tablename in self.analyze_threads and self.analyze_threads[tablename].isAlive())

    def drop_btable(self, tablename):
        """Delete table by tablename."""
        if self.is_analyzing(tablename):
            self.cancel_analyze(tablename)
            raise utils.BayesDBError('Error: cannot drop btable with ANALYZE in progress. Attempting to cancel ANALYZE; please retry drop btable once ANALYZE is successfully completed.')
        self.persistence_layer.drop_btable(tablename)
        return dict()

    def list_btables(self):
        """Return names of all btables."""
        return dict(column_labels=['btable'], data=[[name] for name in self.persistence_layer.list_btables()])

    def label_columns(self, tablename, mappings):
        """
        Add column labels to table in persistence layer, replacing
        labels without warning. Mappings is a dict of column names
        and their labels as given by the user.
        No length is enforced on labels - should we?
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        # Get column names for table so labels don't get written for nonexistent columns.
        M_c_full = self.persistence_layer.get_metadata_full(tablename)['M_c_full']
        colnames_full = utils.get_all_column_names_in_original_order(M_c_full)

        labels_edited = {}
        # Only add labels or overwrite one-by-one.
        for colname, label in mappings.items():
            if colname in colnames_full:
                self.persistence_layer.add_column_label(tablename, colname, label)
                labels_edited[colname] = label
            else:
                raise utils.BayesDBColumnDoesNotExistError(colname, tablename)

        labels = self.persistence_layer.get_column_labels(tablename)
        ret = {'data': [[c, l] for c, l in labels_edited.items()], 'column_labels': ['column', 'label']}
        ret['message'] = "Updated column labels for %s." % (tablename)
        return ret

    def update_codebook(self, tablename, codebook):
            """ Adds a codebook to a btable from a .csv

            Args:
                    tablename (str): Name of the btable
                    codebook (list of dict): The processed codebook. Should come from client because we
                            assume that .csv is stored on the client machine.
            """
            self.persistence_layer.update_codebook(tablename, codebook)

            return dict(message="Updated codebook for %s." % (tablename))

    def update_descriptions(self, tablename, mappings):
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        try:
            M_c_full = self.persistence_layer.get_metadata_full(tablename)['M_c_full']
        except utils.BayesDBError:
            raise utils.BayesDBError("Error: DESCRIBE found no metadata_full file. This is most likely a result of this btable being created with an old version of BayesDB. Please try recreating the table from the original csv, and loading any models you might have.")

        if 'column_codebook' not in    M_c_full.keys():
            raise utils.BayesDBError("Error: DESCRIBE found no codebook in metadata_full file. Please add a codebook.")

        descriptions_edited = dict()
        for colname, description in mappings.items():
            col_idx = M_c_full['name_to_idx'][colname]
            M_c_full['column_codebook'][col_idx]['description'] = description
            descriptions_edited[colname] = description

        self.persistence_layer.update_metadata_full(tablename, M_c_full=M_c_full)

        ret = {
            'data': [[c,d] for c, d in descriptions_edited.items() ],
            'column_labels': ['column', 'description']
        }
        ret['message'] = "Updated column descriptions for %s." % (tablename)

        return ret

    def update_short_names(self, tablename, mappings):
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        try:
            M_c_full = self.persistence_layer.get_metadata_full(tablename)['M_c_full']
        except utils.BayesDBError:
            raise utils.BayesDBError("Error: DESCRIBE found no metadata_full file. This is most likely a result of this btable being created with an old version of BayesDB. Please try recreating the table from the original csv, and loading any models you might have.")

        if 'column_codebook' not in M_c_full.keys():
            raise utils.BayesDBError("Error: DESCRIBE found no codebook in metadata_full file. Please add a codebook.")

        short_names_edited = dict()
        for colname, short_name in mappings.items():
            col_idx = M_c_full['name_to_idx'][colname]
            M_c_full['column_codebook'][col_idx]['short_name'] = short_name
            short_names_edited[colname] = short_name

        self.persistence_layer.update_metadata_full(tablename, M_c_full=M_c_full)

        ret = {
            'data': [[c,s] for c, s in short_names_edited.items() ],
            'column_labels': ['column', 'short name']
        }
        ret['message'] = "Updated column short names for %s." % (tablename)

        return ret

    def describe(self, tablename, columnset):
        """Show column labels, human-readible name, description, and value metadat (in applicable)"""
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        labels = self.persistence_layer.get_column_labels(tablename)

        M_c = self.persistence_layer.get_metadata(tablename)['M_c']

        if columnset == None:
            colnames = M_c['name_to_idx'].keys()
        elif columnset.asList()[0] == '*':
            colnames = M_c['name_to_idx'].keys()
        else:
            column_lists = self.persistence_layer.get_column_lists(tablename)
            colnames = utils.process_column_list(columnset.asList(), M_c, column_lists, dedupe=True)
            colnames = [c.lower() for c in colnames]

        try:
            M_c_full = self.persistence_layer.get_metadata_full(tablename)['M_c_full']
        except utils.BayesDBError:
            raise utils.BayesDBError("Error: DESCRIBE found no metadata_full file. This is most likely a result of this btable being created with an old version of BayesDB. Please try recreating the table from the original csv, and loading any models you might have.")

        ret_table = []
        for c in colnames:
            row = [c]
            colname_idx_M_c = M_c_full['name_to_idx'][c]
            try:
                    colname_codebook = M_c_full['column_codebook'][colname_idx_M_c]
                    column_metadata = M_c_full['column_metadata'][colname_idx_M_c]
            except IndexError:
                    raise utils.BayesDBError("Error: DESCRIBE cannot locate codebook in metatdata. Please add "
                                                                        "a codebook to your table.")

            if colname_codebook is None:
                    raise utils.BayesDBError("Error: DESCRIBE cannot locate codebook in metatdata. Please add "
                                                                        "a codebook to your table.")

            if colname_codebook['description']:
                    short_name = colname_codebook['short_name']
                    description = colname_codebook['description']
            else:
                    short_name = c
                    description = "No description"

            # for each column pull the data info and the description and put it in the table
            data_info_string = ""
            if column_metadata['parameters']:
                    data_info_string += str(column_metadata['parameters']) + "\n"

            if column_metadata['code_to_value']:
                    data_info_string += str(column_metadata['value_to_code'].values()) + "\n"

            ret_table.append([c, short_name, description, data_info_string])

            column_labels = ['column', 'label', 'decription', 'data summary']

        ret = {'data': ret_table, 'column_labels': column_labels}
        ret['message'] = "Describing for %s." % (tablename)
        return ret

    def show_labels(self, tablename, columnset):
        """
        Show column labels for the columns in columnstring
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        labels = self.persistence_layer.get_column_labels(tablename)

        # Get colnames from columnstring
        if columnset == None:
            colnames = labels.keys()
        else:
            column_lists = self.persistence_layer.get_column_lists(tablename)
            M_c = self.persistence_layer.get_metadata(tablename)['M_c']
            colnames = utils.process_column_list(columnset.asList(), M_c, column_lists, dedupe=True)
            colnames = [c.lower() for c in colnames]

        ret = {'data': [[c, l] for c, l in labels.items() if c in colnames], 'column_labels': ['column', 'label']}
        ret['message'] = "Showing labels for %s." % (tablename)
        return ret

    def update_metadata(self, tablename, mappings):
        """
        Add user metadata to table in persistence layer, replacing
        values without warning. Mappings is a dict of key names
        and their values as given by the user.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        for key, value in mappings.items():
                self.persistence_layer.add_user_metadata(tablename, key, value)

        metadata = self.persistence_layer.get_user_metadata(tablename)
        ret = {'data': [[k, v] for k, v in metadata.items() if k in mappings.keys()], 'column_labels': ['key', 'value']}
        ret['message'] = "Updated user metadata for %s." % (tablename)
        return ret

    def show_metadata(self, tablename, keyset):
        """
        Get user metadata from persistence layer and show the values for the keys specified
        by the user. If no keystring is given, show all metadata key-value pairs.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        metadata = self.persistence_layer.get_user_metadata(tablename)
        if keyset == None:
            metadata_keys = metadata.keys()
        else:
            metadata_keys = keyset.asList()

        ret = {'data': [[k, metadata[k]] for k in metadata_keys if k in metadata], 'column_labels': ['key', 'value']}
        ret['message'] = "Showing user metadata for %s." % (tablename)
        return ret

    def update_schema(self, tablename, mappings):
        """
        mappings is a dict of column name to 'numerical', 'categorical', 'cyclic', or 'ignore'
        Cannot update type for key column, or change another column to key.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        if self.persistence_layer.has_models(tablename):
            raise utils.BayesDBError("Error: btable %s already has models. The schema may not be updated after models have been initialized; please either create a new btable or drop the models from this one." % tablename)

        msg = self.persistence_layer.update_schema(tablename, mappings)
        ret = self.show_schema(tablename)
        ret['message'] = 'Updated schema.'
        return ret

    def create_btable_from_existing(self, tablename, query_colnames, cctypes_existing_full, M_c_existing_full, query_data):
        """
        Used in INTO statements to create a new btable as a portion of an existing one.
        """
        ## First, test if table with this name already exists, and fail if it does
        if self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBError('Btable with name %s already exists.' % tablename)

        cctypes_full, warnings = data_utils.guess_column_types(query_data, query_colnames)
        for query_idx, query_colname in enumerate(query_colnames):
                if query_colname in M_c_existing_full['name_to_idx']:
                        cctypes_full[query_idx] = cctypes_existing_full[M_c_existing_full['name_to_idx'][query_colname]]

        if 'key' not in cctypes_full:
                raw_T_full, colnames_full, cctypes_full = data_utils.select_key_column(query_data, query_colnames, cctypes_full, key_column=None, testing=self.testing)
        else:
                raw_T_full = query_data
                colnames_full = query_colnames
        T_full, M_r_full, M_c_full, _ = data_utils.gen_T_and_metadata(colnames_full, raw_T_full, cctypes=cctypes_full)

        # variables without "_full" don't include ignored columns.
        raw_T, cctypes, colnames = data_utils.remove_ignore_cols(T_full, cctypes_full, colnames_full)
        T, M_r, M_c, _ = data_utils.gen_T_and_metadata(colnames, raw_T, cctypes=cctypes)
        self.persistence_layer.create_btable(tablename, cctypes_full, cctypes, T, M_r, M_c, T_full, M_r_full, M_c_full, query_data)

        return dict(column_labels=colnames_full, data=[cctypes_full], message='Created btable %s. Schema taken from original btable:' % tablename, warnings=warnings)

    def create_btable(self, tablename, header, raw_T_full, cctypes_full=None, key_column=None, subsample=False, codebook=None):
        """
        Upload a csv table to the predictive db.
        cctypes must be a dictionary mapping column names
        to either 'ignore', 'numerical', 'categorical', or 'cyclic'. Not every
        column name must be present in the dictionary: default is numerical.

        subsample is False by default, but if it is passed an int, it will subsample using
        <subsample> rows for the initial ANALYZE, and then insert all other rows afterwards.
        """
        ## First, test if table with this name already exists, and fail if it does
        if self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBError('Btable with name %s already exists.' % tablename)

        warnings = []

        # variables with "_full" include ignored columns.
        colnames_full = [h.lower().strip() for h in header]
        raw_T_full = data_utils.convert_nans(raw_T_full)

        if cctypes_full is None:
            cctypes_full, warnings = data_utils.guess_column_types(raw_T_full, colnames_full)

        raw_T_full, colnames_full, cctypes_full = data_utils.select_key_column(raw_T_full, colnames_full, cctypes_full, key_column, testing=self.testing)
        T_full, M_r_full, M_c_full, _ = data_utils.gen_T_and_metadata(colnames_full, raw_T_full, cctypes=cctypes_full, codebook=codebook)

        # variables without "_full" don't include ignored columns.
        raw_T, cctypes, colnames = data_utils.remove_ignore_cols(raw_T_full, cctypes_full, colnames_full)
        T, M_r, M_c, _ = data_utils.gen_T_and_metadata(colnames, raw_T, cctypes=cctypes, codebook=codebook)

        # if subsampling enabled, create the subsampled version of T (be sure to use non-subsampled M_r and M_c)
        # T_sub is T with only first <subsample> rows;
        # TODO: in future can randomly pick rows, but need to reorder T post-ANALYZE to achieve T's original order.
        T_sub = None
        if subsample:
            assert type(subsample) == int
            T_sub = T[:subsample]

        self.persistence_layer.create_btable(tablename, cctypes_full, cctypes, T, M_r, M_c, T_full, M_r_full, M_c_full, raw_T_full, T_sub)

        schema = self.show_schema(tablename)

        return dict(column_labels=schema['column_labels'], data=schema['data'], message='Created btable %s. Inferred schema:' % tablename, warnings=warnings)

    def check_btable_created_and_checked(self, tablename, client_online=False):
        """
        Checks whether a given table has been created and checked. If called from the online client,
        returns a dict with 'results' key, otherwise returns a tuple.
        """
        table_created = tablename in self.persistence_layer.btable_index
        table_checked = tablename in self.persistence_layer.btable_check_index

        if client_online:
                return { 'result': {'table_created':table_created, 'table_checked':table_checked} }
        else:
                return table_created, table_checked

    def upgrade_btable(self, tablename, upgrade_key_column=None):
        """
        Btables created in early versions of BayesDB may not have attributes that are required in more
        recent versions of BayesDB (example: required key column). This function allows them to be
        upgraded simply, without having to manually recreate the btable and re-load models.
        """
        # 1. Check for readable metadata file - if not found or not readable, the table's data files
        #             have become corrupted, and the table must be dropped.
        try:
                metadata = self.persistence_layer.get_metadata(tablename)
        except utils.BayesDBError:
                print "Metadata for %s have been corrupted, and the table must be dropped." % tablename
                print "Press any key to confirm, and then re-create it using CREATE BTABLE"
                if not self.testing:
                        user_input = raw_input()
                self.drop_btable(tablename)
                return

        # 2. Check for metadata_full file - if not found, create it from metadata.
        #             Okay to use metadata since btables created that long ago won't have ignore/key columns.
        try:
                metadata_full = self.persistence_layer.get_metadata_full(tablename)
        except utils.BayesDBError:
                metadata = self.persistence_layer.get_metadata(tablename)
                metadata_full = metadata
                # Rename metadata keys with suffix '_full'
                for key in metadata.keys():
                        metadata_full[key + '_full'] = metadata_full.pop(key)
                # Btables created without a metadata_full file don't have raw_T_full saved in metadata, so we need to recreate it.
                metadata_full['raw_T_full'] = data_utils.gen_raw_T_full_from_T_full(metadata_full['T_full'], metadata_full['M_c_full'])
                self.persistence_layer.write_metadata_full(tablename, metadata_full)
                print "Upgraded %s: added metadata_full file" % tablename

        # 3. Check for key column - if not found, add one.
        if 'key' not in metadata_full['cctypes_full']:
                print "Table %s is from an old version of BayesDB and does not have a key column." % tablename
                raw_T_full = metadata_full['raw_T_full']
                colnames_full = utils.get_all_column_names_in_original_order(metadata_full['M_c_full'])
                cctypes_full = metadata_full['cctypes_full']

                raw_T_full, colnames_full, cctypes_full = data_utils.select_key_column(raw_T_full, colnames_full, cctypes_full, testing=self.testing)
                T_full, M_r_full, M_c_full, _ = data_utils.gen_T_and_metadata(colnames_full, raw_T_full, cctypes=cctypes_full)

                # Don't need to update T, M_r, M_c, cctypes (variables without "_full")
                #     because they don't include ignore/key columns.
                self.persistence_layer.upgrade_btable(tablename, cctypes_full, T_full, M_r_full, M_c_full, raw_T_full)
                print "Upgraded %s: added key column" % tablename

        # 4. If cctypes are stored as multinomial or continuous, change to
        #     categorical and numerical, respectively.
        cctypes_full = metadata_full['cctypes_full']
        if 'continuous' in cctypes_full or 'multinomial' in cctypes_full:
                colnames_full = utils.get_all_column_names_in_original_order(metadata_full['M_c_full'])

                mappings = dict()
                for colname, cctype in zip(colnames_full, cctypes_full):
                        if cctype == 'continuous':
                                mappings[colname] = dict(cctype = 'numerical', parameters = None)
                        elif cctype == 'multinomial':
                                mappings[colname] = dict(cctype = 'categorical', parameters = None)

                self.persistence_layer.update_schema(tablename, mappings)
                print "Upgraded %s: converted column types 'multinomial' to 'categorical' and 'continuous' to 'numerical'" % tablename

        self.persistence_layer.btable_check_index.append(tablename)

        # 5. Create a blank codebook, if necessary
        if 'column_codebook' in metadata_full['M_c_full'].keys():
                if metadata_full['M_c_full']['column_codebook'] is None:
                        self.persistence_layer.add_default_codebook_to_metadata(tablename)
                        print("Created empty codebook for %s." % (tablename))
        elif 'column_codebook' not in metadata_full.keys():
                self.persistence_layer.add_default_codebook_to_metadata(tablename)
                print("Createda empty codebook for %s." % (tablename))

    def show_schema(self, tablename):
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        metadata_full = self.persistence_layer.get_metadata_full(tablename)
        M_c_full = metadata_full['M_c_full']
        colnames_full = utils.get_all_column_names_in_original_order(M_c_full)
        parameters_full = [str(x['parameters']) for x in M_c_full['column_metadata']]
        cctypes_full = metadata_full['cctypes_full']

        schema_full = zip(colnames_full, cctypes_full, parameters_full)

        return dict(column_labels=['column', 'datatype', 'parameters'], data=schema_full)

    def save_models(self, tablename):
        """Opposite of load models! Returns the models, including the contents, which
        the client then saves to disk (in a pickle file)."""
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        models = self.persistence_layer.get_models(tablename)
        schema = self.persistence_layer.get_schema(tablename)
        return dict(models=models, schema=schema)

    def load_models(self, tablename, models, model_schema):
        """Load these models as if they are new models"""
        # Models are stored in the format: dict[model_id] = dict[X_L, X_D, iterations].
        # We just want to pass the values.

        # For backwards compatibility with v0.1, where models are stored in the format:
        # dict[X_L_list, X_D_list, M_c, T]
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        if 'X_L_list' in models:
            print """WARNING! The models you are currently importing are stored in an old format
                 (from version 0.1); it is deprecated and may not be supported in future releases.
                 Please use "SAVE MODELS FROM <btable> TO <filename.pkl.gz>" to create an updated copy of your models."""

            old_models = models
            models = dict()
            for id, (X_L, X_D) in enumerate(zip(old_models['X_L_list'], old_models['X_D_list'])):
                models[id] = dict(X_L=X_L, X_D=X_D, iterations=0)

        if model_schema:
            table_schema = self.persistence_layer.get_schema(tablename)
            # If schemas match, add the models
            # If schemas don't match:
            #     If there are models, don't add the new ones (they'll be incompatible)
            #     If there aren't models, add the new ones.
            if table_schema != model_schema:
                if self.persistence_layer.has_models(tablename):
                    raise utils.BayesDBError("""
                        Table %s already has models under a different schema than
                        the models you are loading. All models used must have the
                        same schema. To load these models, first drop the existing
                        models from the table.
                        """ % tablename)
                else:
                    # Update 'continuous' and 'multinomial', if they exist in model_schema
                    upgrade_map = dict(continuous = 'numerical', multinomial = 'categorical')
                    for colname, mapping in model_schema.items():
                        cctype = mapping['cctype']
                        if cctype in upgrade_map:
                            print 'Converting model schema column type %s to %s' % (cctype, upgrade_map[cctype])
                            model_schema[colname]['cctype'] = upgrade_map[cctype]
                    self.persistence_layer.update_schema(tablename, model_schema)

        result = self.persistence_layer.add_models(tablename, models.values())

        return self.show_models(tablename)

    def drop_models(self, tablename, model_indices=None):
        """Drop the specified models. If model_ids is None or all, drop all models."""
        if self.is_analyzing(tablename):
            raise utils.BayesDBError('Error: cannot drop models with ANALYZE in progress. Please retry once ANALYZE is successfully completed or canceled.')
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        self.persistence_layer.drop_models(tablename, model_indices)
        return self.show_models(tablename)

    def initialize_models(self, tablename, n_models=16, model_config=None):
        """
        Initialize n_models models.

        By default, model_config specifies to use the CrossCat model. You may pass 'naive bayes'
        or 'crp mixture' to use those specific models instead. Alternatively, you can pass a custom
        dictionary for model_config, as long as it contains a kernel_list, initializaiton, and
        row_initialization.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        if self.is_analyzing(tablename):
            raise utils.BayesDBError('Error: cannot initialize models with ANALYZE in progress. Please retry once ANALYZE is successfully completed or canceled.')

        # Get t, m_c, and m_r
        metadata = self.persistence_layer.get_metadata(tablename)
        M_c = metadata['M_c']
        M_r = metadata['M_r']
        if 'T_sub' in metadata and metadata['T_sub']:
            T = metadata['T_sub']
        else:
            T = metadata['T']

        # Set model configuration parameters.
        if type(model_config) == str and model_config.lower() == 'naive bayes':
            model_config = dict(kernel_list=['column_hyperparameters'],
                                initialization='together',
                                row_initialization='together')
        elif type(model_config) == str and model_config.lower() == 'crp mixture':
            model_config = dict(kernel_list=['column_hyperparameters',
                                             'row_partition_hyperparameters',
                                             'row_partition_assignments'],
                                initialization='together',
                                row_initialization='from_the_prior')
        elif type(model_config) != dict or ('kernel_list' not in model_config) or ('initialization' not in model_config) or ('row_initialization' not in model_config):
            # default model_config: crosscat
            model_config = dict(kernel_list=(), # uses default
                                initialization='from_the_prior',
                                row_initialization='from_the_prior')
        else:
            raise utils.BayesDBError("Invalid model config")

        # Make sure the model config matches existing model config, if there are other models.
        existing_model_config = self.persistence_layer.get_model_config(tablename)
        if existing_model_config is not None and existing_model_config != model_config:
            raise utils.BayesDBError("Error: model config must match existing model config: %s" % str(existing_model_config))

        # Call initialize on backend
        init_dict = dict(M_c=M_c, M_r=M_r, T=T, n_chains=n_models,
                         initialization=model_config['initialization'],
                         row_initialization=model_config['row_initialization'])
        X_L_list, X_D_list = self.call_backend('initialize', init_dict)

        # If n_models is 1, initialize returns X_L and X_D instead of X_L_list and X_D_list
        if n_models == 1:
            X_L_list = [X_L_list]
            X_D_list = [X_D_list]

        model_list = list()
        for X_L, X_D in zip(X_L_list, X_D_list):
            # Note: diagnostics entry is created on first collection of diagnostics
            model_list.append(dict(X_L=X_L, X_D=X_D, iterations=0, time=0.0,
                                  model_config=model_config))

        # Insert results into persistence layer
        self.persistence_layer.add_models(tablename, model_list)
        return self.show_models(tablename)

    def show_models(self, tablename):
        """Return the current models and their iteration counts."""
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        model_data = self.persistence_layer.get_models(tablename)
        if 'schema' in model_data.keys():
                models = model_data['models']
        else:
                models = model_data
        modelid_iteration_info = list()
        for modelid, model in sorted(models.items(), key=lambda t:t[0]):
            modelid_iteration_info.append((modelid, model['iterations']))
        if len(models) == 0:
            raise utils.BayesDBNoModelsError(tablename)
        else:
            return dict(models=modelid_iteration_info)

    def show_diagnostics(self, tablename):
        """
        Display diagnostic information for all your models.
        TODO: generate plots of num_views, column_crp_alpha, logscore, and f_z stuff
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        models = self.persistence_layer.get_models(tablename)

        if len(models) == 0:
            raise utils.BayesDBError("No models for btable %s. Create some with the INITIALIZE MODELS command." % tablename)

        if 'diagnostics' not in models[0].keys():
            raise utils.BayesDBError("No diagnostics found. Run ANALYZE to collect some.")

        if 'model_config' not in models[0]:
                raise utils.BayesDBError("The models in %s were created with an old version of BayesDB which does not support model diag_utils." % tablename)

        data = list()
        diagnostic_keys = sorted(models[0]['diagnostics'].keys())
        time_key_index = diagnostic_keys.index('time')
        del diagnostic_keys[time_key_index]
        iterations_key_index = diagnostic_keys.index('iterations')
        del diagnostic_keys[iterations_key_index]

        for modelid, model in sorted(models.items(), key=lambda t:t[0]):
            data_row = [model['diagnostics'][key][-1] for key in diagnostic_keys]
            # removed lengthy str(model['model_config']) at end of row
            data.append(tuple([modelid, model['iterations'], model['time']] + data_row))
        else:
            return dict(column_labels=['model_id', 'iterations', 'time'] + diagnostic_keys, data=data)

    def analyze(self, tablename, model_indices=None, iterations=None, seconds=None, ct_kernel=0,
                background=True):
        """
        Run analyze for the selected table. model_indices may be 'all' or None to indicate all models.

        Runs for a maximum of iterations, or a maximum number of seconds.

        Previously: this command ran in the same thread as this engine.
        Now: runs each model in its own thread, and does 10 seconds of inference at a time,
        by default. Each thread also has its own crosscat engine instance!

        Args:
            tablename (type): BayesDB btable name
            iterations (int): Number of iterations to run.
            seconds (int): Number of seconds to run.
            ct_kernel (int): Specifies which crosscat column transition kernel to run. 0: Gibbs,
                1: Metropolis-Hastings
            background (bool): Run asynchronously.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        if not self.persistence_layer.has_models(tablename):
            raise utils.BayesDBNoModelsError(tablename)

        # Get t, m_c, and m_r, and tableid. Use subsampled T if subsampling is enabled.
        metadata = self.persistence_layer.get_metadata(tablename)
        M_c = metadata['M_c']
        M_r = metadata['M_r']
        T = metadata['T']
        if 'T_sub' in metadata:
            T_sub = metadata['T_sub']
        else:
            T_sub = None

        max_model_id = self.persistence_layer.get_max_model_id(tablename)
        if max_model_id == -1:
            raise utils.BayesDBError("You must INITIALIZE MODELS before using ANALYZE.")
        models = self.persistence_layer.get_models(tablename)

        if model_indices is None or (str(model_indices).upper() == 'ALL'):
            modelids = sorted(models.keys())
        else:
            assert type(model_indices) == list
            modelids = model_indices

        first_model = models[modelids[0]]
        if 'model_config' in first_model and 'kernel_list' in first_model['model_config']:
            kernel_list = first_model['model_config']['kernel_list']
        else:
            kernel_list = () # default kernel list


        if tablename in self.analyze_threads and self.analyze_threads[tablename].isAlive():
            raise utils.BayesDBError("{0} is already being analyzed. Try using 'SHOW ANALYZE FOR {0}' to get more information, or 'CANCEL ANALYZE FOR {0}' to cancel the ANALYZE.".format(tablename))


        # Start analyze thread.
        t = AnalyzeMaster(args=(tablename, modelids, kernel_list, iterations,
                                                        seconds, M_c, T, T_sub, models, background, self))
        self.analyze_threads[tablename] = t
        t.start()

        if not background:
            t.join() # just wait for the AnalyzeMaster thread to finish before returning.
            return dict(message="Analyze complete.")
        else:
            return dict(message="Analyzing %s: models will be updated in the background." % tablename)

    def _insert_subsampled_rows(self, tablename, T, T_sub, M_c, X_L_list, X_D_list, modelids):
        """
        If the analyze in progress was using subsampling, then insert all the non-subsampled rows now.
        TODO: asserts are costly, remove before performance-critical use. They're just there for code
        correctness and readability right now.
        """
        assert T[:len(T_sub)] == T_sub
        new_rows = T[len(T_sub):]
        insert_args = dict(M_c=M_c, T=T_sub, X_L_list=X_L_list, X_D_list=X_D_list, new_rows=new_rows)
        X_L_list, X_D_list, T_new = self.call_backend('insert', insert_args)
        assert T_new == T

        # TODO: each loop through this for loop is parallelizable
        for i, modelid in enumerate(modelids):
            self.persistence_layer.update_model(tablename, X_L_list[i], X_D_list[i], dict(), modelid)


    def show_analyze(self, tablename):
        if tablename in self.analyze_threads and self.analyze_threads[tablename].isAlive():
            info = self.analyze_threads[tablename].get_info()
            return dict(message=info)
        else:
            return dict(message="Table %s is not currently being analyzed." % tablename)

    def cancel_analyze(self, tablename):
        if tablename in self.analyze_threads and self.analyze_threads[tablename].isAlive():
            self.analyze_threads[tablename].stop()
            return dict(message="Analyze for table %s is being canceled. Please wait a few moments for it to exit safely." % tablename)
        else:
            raise utils.BayesDBError("Table %s is not being analyzed." % tablename)


    def infer(self, tablename, functions, confidence, whereclause, limit, numsamples, order_by=False, plot=False, modelids=None, summarize=False, hist=False, freq=False, newtablename=None):
        """
        Impute missing values.
        Sample INFER: INFER columnstring FROM tablename WHERE whereclause WITH confidence LIMIT limit;
        Sample INFER INTO: INFER columnstring FROM tablename WHERE whereclause WITH confidence INTO newtablename LIMIT limit;
        Argument newtablename == null/emptystring if we don't want to do INTO

        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        if not self.persistence_layer.has_models(tablename):
            raise utils.BayesDBNoModelsError(tablename)

        if numsamples is None:
            numsamples=50 ##TODO: put this in a config file

        return self.select(tablename, functions, whereclause, limit, order_by,
                                             impute_confidence=confidence, numsamples=numsamples, plot=plot, modelids=modelids, summarize=summarize, hist=hist, freq=freq, newtablename=newtablename)

    def select(self, tablename, functions, whereclause, limit, order_by, impute_confidence=None, numsamples=None, plot=False, modelids=None, summarize=False, hist=False, freq=False, newtablename=None):
        """
        BQL's version of the SQL SELECT query.

        First, reads codes from T and converts them to values.
        Then, filters the values based on the where clause.
        Then, fills in all imputed values, if applicable.
        Then, orders by the given order_by functions.
        Then, computes the queried values requested by the column string.

        query_colnames is the list of the raw columns/functions from the columnstring, with row_id prepended
        queries is a list of (query_function, query_args, aggregate) tuples, where 'query_function' is
            a function like row_id, column, similarity, or typicality, and 'query_args' are the function-specific
            arguments that that function takes (in addition to the normal arguments, like M_c, X_L_list, etc).
            aggregate specifies whether that individual function is aggregate or not

        queries: [(func, f_args, aggregate)]
        order_by: [(func, f_args, desc)]
        where_conditions: [(func, f_args, op, val)]

        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        M_c_full, M_r_full, T_full = self.persistence_layer.get_metadata_and_table_full(tablename)

        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        column_lists = self.persistence_layer.get_column_lists(tablename)

        key_column_name = self.persistence_layer.get_key_column_name(tablename)

        # Parse queries, where_conditions, and order by.
        queries, query_colnames = self.parser.parse_functions(functions, M_c, T, M_c_full, T_full, column_lists, key_column_name)
        if whereclause == None:
            where_conditions = []
        else:
            where_conditions = self.parser.parse_where_clause(whereclause, M_c, T, M_c_full, T_full, column_lists)
            if len(where_conditions) > 0:
                assert len(where_conditions[0]) == 4
        if order_by == False:
            order_by = []
        else:
            order_by = self.parser.parse_order_by_clause(order_by, M_c, T, M_c_full, T_full, column_lists)

        # Fill in individual impute confs with impute_confidence if they're None and impute_confidence isn't:
        if impute_confidence is not None:
            for f_list in (where_conditions, queries, order_by):
                for i in range(len(f_list)):
                    if f_list[i][0] == funcs._column and f_list[i][1][1] is None:
                        new_f = list(f_list[i]) # (function, args, ...)
                        new_f_args = list(new_f[1]) # args = (idx, conf)
                        new_f_args[1] = impute_confidence
                        new_f[1] = tuple(new_f_args)
                        f_list[i] = tuple(new_f)

        if len(X_L_list) == 0:
            select_utils.check_if_functions_need_models(queries, tablename, order_by, where_conditions)

        ## Construct function list
        query_functions = [(q[0], q[1]) for q in queries]
        order_by_functions = [(o[0], o[1]) for o in order_by]
        where_functions = [(w[0], w[1]) for w in where_conditions]

        function_list = [q for q in query_functions]

        # Look up index in the function list, based on your own index.
        query_idxs = range(len(query_functions))
        order_by_idxs = []
        where_idxs = []

        query_size = len(query_functions)
        order_by_size = query_size

        def get_existing_f_idx(new_fun, function_list):
                for f_idx, fun in enumerate(function_list):
                    if new_fun == fun:
                        return f_idx
                return False

        for o_idx, o_fun in enumerate(order_by_functions):
            existing_f_idx = get_existing_f_idx(o_fun, function_list)
            if existing_f_idx:
                order_by_idxs.append(existing_f_idx)
            else:
                function_list.append(o_fun)
                order_by_idxs.append(len(function_list) - 1)
                order_by_size += 1

        for w_idx, w_fun in enumerate(where_functions):
            existing_f_idx = get_existing_f_idx(w_fun, function_list)
            if existing_f_idx:
                where_idxs.append(existing_f_idx)
            else:
                function_list.append(w_fun)
                where_idxs.append(len(function_list) - 1)


        ## Now function list is constructed, so create the data table one row at a time,
        ## applying where clause in the process.
        ## data_tuples means row_id, row tuples
        data_tuples = []
        for row_id, T_row in enumerate(T):
            row = [None]*order_by_size
            row_values = select_utils.convert_row_from_codes_to_values(T_row, M_c)
            where_clause_eval = select_utils.evaluate_where_on_row(row_id, row_values, where_conditions, M_c, M_c_full,
                                                             X_L_list, X_D_list, T, T_full, self, tablename, numsamples, impute_confidence)
            if type(where_clause_eval) == list:
                for w_idx, val in enumerate(where_clause_eval):
                    # Now, store the already-computed values, if they're used for anything besides the where clause.
                    if where_idxs[w_idx] < order_by_size:
                        row[where_idxs[w_idx]] = val
                data_tuples.append((row_id, row))

        ## Now, order the rows. First compute all values that will be necessary for order by.
        if len(order_by) > 0:
            scored_data_tuples = list() # Entries are score, data_tuple
            for (row_id, row) in data_tuples:
                scores = []
                for o_idx, (f, args, desc) in enumerate(order_by):
                    # Look up whether function evaluated earlier in select. # TODO: can functions ever return None?
                    if row[order_by_idxs[o_idx]] is not None:
                        score = row[order_by_idxs[o_idx]]
                    else:
                        if f != funcs._column_ignore:
                            score = f(args, row_id, row, M_c, X_L_list, X_D_list, T, self, numsamples)
                        else:
                            score = f(args, row_id, row, M_c_full, T_full, self)
                        # Save value, if it will be displayed in output.
                        if order_by_idxs[o_idx] < query_size:
                            row[order_by_idxs[o_idx]] = score

                    if desc:
                        if data_utils.get_can_cast_to_float([score]):
                            score = float(score)
                        score *= -1
                    scores.append(score)
                scored_data_tuples.append((tuple(scores), (row_id, row)))
            scored_data_tuples.sort(key=lambda tup: tup[0], reverse=False)
            data_tuples = [(tup[1][0], tup[1][1][:query_size]) for tup in scored_data_tuples]

        ## Now, apply the limit and compute the queried columns.

        # Compute aggregate functions just once, then cache them.
        aggregate_cache = dict()
        for q_idx, (query_function, query_args, aggregate) in enumerate(queries):
            if aggregate:
                aggregate_cache[q_idx] = query_function(query_args, None, None, M_c, X_L_list, X_D_list, T, self, numsamples)

        # Only return one row if all aggregate functions (row_id will never be aggregate, so subtract 1 and don't return it).
        if len(aggregate_cache) == len(queries):
            limit = 1

        # Iterate through data table, calling each query_function to fill in the output values.
        data = []
        row_count = 0
        for row_id, row in data_tuples:
            for q_idx, (query_function, query_args, aggregate) in enumerate(queries):
                if aggregate:
                    row[q_idx] = aggregate_cache[q_idx]
                elif row[q_idx] is None:
                    if query_function != funcs._column_ignore:
                        row[q_idx] = query_function(query_args, row_id, row, M_c, X_L_list, X_D_list, T, self, numsamples)
                    else:
                        row[q_idx] = query_function(query_args, row_id, row, M_c_full, T_full, self)
            data.append(tuple(row))
            row_count += 1
            if row_count >= limit:
                break

        # Execute INTO statement
        if newtablename is not None:
            cctypes_full = self.persistence_layer.get_cctypes_full(tablename)
            self.create_btable_from_existing(newtablename, query_colnames, cctypes_full, M_c_full, data)

        if summarize | hist | freq:
            if summarize:
                data, columns = utils.summarize_table(data, query_colnames, M_c, remove_key=True)
            elif hist:
                data, columns = utils.histogram_table(data, query_colnames, M_c, remove_key=True)
            elif freq:
                data, columns = utils.freq_table(data, query_colnames, M_c, remove_key=True)
            column_names = columns
        else:
                columns = data_utils.get_column_labels_from_M_c(M_c, query_colnames)
                column_names = query_colnames
        ret = dict(data = data, column_labels = columns, column_names = column_names)
        if plot:
            ret['M_c'] = M_c
            ret['schema_full'] = self.persistence_layer.get_schema_full(tablename)

            try:
                    ret['metadata_full'] = self.persistence_layer.get_metadata_full(tablename)
            except BayesDBError:
                    pass

        return ret


    def simulate(self, tablename, functions, givens, numpredictions, order_by, plot=False, modelids=None, summarize=False, hist=False, freq=False, newtablename=None):
        """Simple predictive samples. Returns one row per prediction, with all the given and predicted variables."""
        # TODO: whereclause not implemented.
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        if not self.persistence_layer.has_models(tablename):
            raise utils.BayesDBNoModelsError(tablename)

        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        if len(X_L_list) == 0:
            return {'message': 'You must INITIALIZE MODELS (and highly preferably ANALYZE) before using predictive queries.'}
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)

        numrows = len(M_r['idx_to_name'])
        name_to_idx = M_c['name_to_idx']

        ## TODO throw exception for <,> without dissallowing them in the values.
        given_col_idxs_to_vals = dict()
        Y = None
        if givens is not None:
            Y = []
            for given_condition in givens.given_conditions:
                column_value = given_condition.value
                column_name = given_condition.column
                column_value = utils.string_to_column_type(column_value, column_name, M_c)
                given_col_idxs_to_vals[name_to_idx[column_name]] = column_value
                Y.append((numrows+1, name_to_idx[column_name], column_value))

            # map values to codes
            Y = [(r, c, data_utils.convert_value_to_code(M_c, c, colval)) for r,c,colval in Y]

        ## Parse queried columns.
        column_lists = self.persistence_layer.get_column_lists(tablename)
        # Set M_c_full to None because we don't want to simulate key/ignore columns
        queries, query_colnames = self.parser.parse_functions(functions, M_c, T, M_c_full=None, T_full=None, column_lists=column_lists)
        ##TODO check duplicates
        ##TODO check for no functions

        ##TODO col_indices, colnames are a hack from old parsing

        col_indices = [query[1][0] for query in queries]
        query_col_indices = [idx for idx in col_indices if idx not in given_col_idxs_to_vals.keys()]
        Q = [(numrows+1, col_idx) for col_idx in query_col_indices]

        if len(Q) > 0:
            out = self.call_backend('simple_predictive_sample', dict(M_c=M_c, X_L=X_L_list, X_D=X_D_list, Y=Y, Q=Q, n=numpredictions))
        else:
            out = [[] for x in range(numpredictions)]
        assert type(out) == list and len(out) >= 1 and type(out[0]) == list and len(out) == numpredictions

        # convert to data, columns dict output format
        # map codes to original values
        data = []
        for out_row in out:
            row = []
            i = 0
            for idx in col_indices:
                if idx in given_col_idxs_to_vals:
                    row.append(given_col_idxs_to_vals[idx])
                else:
                    row.append(data_utils.convert_code_to_value(M_c, idx, out_row[i]))
                    i += 1
            data.append(row)

        # Execute INTO statement
        if newtablename is not None:
            cctypes = self.persistence_layer.get_cctypes(tablename)
            newtable_info = self.create_btable_from_existing(newtablename, query_colnames, cctypes, M_c, data)
            if 'key' in query_colnames:
                    query_colnames.remove('key')

        if summarize | hist | freq:
            if summarize:
                data, columns = utils.summarize_table(data, query_colnames, M_c, remove_key=False)
            elif hist:
                data, columns = utils.histogram_table(data, query_colnames, M_c, remove_key=False)
            elif freq:
                data, columns = utils.freq_table(data, query_colnames, M_c, remove_key=False)
            column_names = columns
        else:
                columns = data_utils.get_column_labels_from_M_c(M_c, query_colnames)
                column_names = query_colnames
        ret = dict(data = data, column_labels = columns, column_names = column_names)
        if plot:
            ret['M_c'] = M_c
            ret['schema_full'] = self.persistence_layer.get_schema_full(tablename)

        return ret

    def show_column_lists(self, tablename):
        """
        Return a list of all column list names.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        column_lists = self.persistence_layer.get_column_lists(tablename)
        return dict(column_labels=['column list'], data=[[k] for k in column_lists.keys()])

    def show_row_lists(self, tablename):
        """
        Return a list of all row lists, and their row counts.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        row_lists = self.persistence_layer.get_row_lists(tablename)
        return dict(column_labels=['Row List Name', 'Row Count'], data=[(name, len(rows)) for (name, rows) in row_lists.items()])


    def show_columns(self, tablename, column_list=None):
        """
        Return the specified columnlist. If None, return all columns in original order.
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        if column_list:
            column_names = self.persistence_layer.get_column_list(tablename, column_list)
        else:
            M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
            column_names = list(M_c['name_to_idx'].keys())
        return dict(column_labels=column_names)

    def show_model(self, tablename, modelid, filename):
        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        import crosscat.utils.plot_utils
        crosscat.utils.plot_utils.plot_views(numpy.array(T), X_D_list[modelid], X_L_list[modelid], M_c, filename)

    def create_column_list(self, tablename, functions, name):
        """
        Create the column list with the specified column names (functions).
        """
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        column_lists = self.persistence_layer.get_column_lists(tablename)
        queries, column_names = self.parser.parse_functions(functions, M_c, T, M_c, T, column_lists, key_column_name=None)

        # save column list, if given a name to save as
        if name:
            self.persistence_layer.add_column_list(tablename, name, column_names)

        return dict(column_labels=column_names)


    def estimate_columns(self, tablename, functions, whereclause, limit, order_by, name=None, modelids=None, numsamples=None):
        """
        Return all the column names from the specified table as a list.
        First, columns are filtered based on whether they match the whereclause.
        The whereclause must consist of functions of a single column only.
        Next, the columns are ordered by other functions of a single column.
        Finally, the columns are limited to the specified number.

        ## allowed functions:
        # typicality(centrality)
        # dependence probability to <col>
        # mutual information with <col>
        # correlation with <col>
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)

        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        column_indices = list(M_c['name_to_idx'].values())

        # Store the names of each of the functions in the where or order by, to be displayed in output.
        func_descriptions = []

        ## filter based on where clause
        where_conditions = self.parser.parse_column_whereclause(whereclause, M_c, T)
        if where_conditions is not None and len(X_L_list) == 0:
            raise utils.BayesDBNoModelsError(tablename)
        column_indices_and_data = estimate_columns_utils.filter_column_indices(column_indices, where_conditions, M_c, T, X_L_list, X_D_list, self, numsamples)
        func_descriptions += [estimate_columns_utils.function_description(where_item[0][0], where_item[0][1], M_c) for where_item in where_conditions]

        ## order
        if order_by and len(X_L_list) == 0:
            raise utils.BayesDBNoModelsError(tablename)
        if order_by != False:
            order_by = self.parser.parse_column_order_by_clause(order_by, M_c)
            func_descriptions += [estimate_columns_utils.function_description(order_item[0], order_item[1], M_c) for order_item in order_by]

        # Get tuples of column estimation function values and column indices
        ordered_tuples = estimate_columns_utils.order_columns(column_indices_and_data, order_by, M_c, X_L_list, X_D_list, T, self, numsamples)
        # tuple contains: (tuple(data), cidx)

        # limit
        if limit != float('inf'):
            ordered_tuples = ordered_tuples[:limit]

        # convert indices to names and create data list of column names and associated function values
        column_names = []
        column_data = []

        # column_data should be a list of the data values
        for ordered_tuple in ordered_tuples:
            if type(ordered_tuple) == int:
                #
                ordered_tuple = ((), column_idx_tup)
            column_name_temp = M_c['idx_to_name'][str(ordered_tuple[1])]
            column_names.append(column_name_temp)

            # Get the values of the functions that were used in where and order by
            column_data_temp = [data_utils.get_column_labels_from_M_c(M_c, [column_name_temp])[0]]
            column_data_temp.append(column_name_temp)
            column_data_temp.extend(ordered_tuple[0])
            column_data.append(column_data_temp)

        # save column list, if given a name to save as
        if name:
            self.persistence_layer.add_column_list(tablename, name, column_names)

        # De-duplicate values, if the same function was used in where and order by
        used_names = set()
        indices_to_pop = []
        func_descriptions.insert(0, 'column label')
        func_descriptions.insert(1, 'column name')
        for i, name in enumerate(func_descriptions):
            if name not in used_names:
                used_names.add(name)
            else:
                indices_to_pop.append(i)
        indices_to_pop.sort(reverse=True) # put highest indices first, so popping them doesn't affect others
        for i in indices_to_pop:
            func_descriptions.pop(i)
        for column_data_list in column_data:
            for i in indices_to_pop:
                column_data_list.pop(i)


        # Create column names ('column' followed by a string describing each function)
        ret = dict()
        if column_data != []:
            ret['column_labels'] = func_descriptions
            ret['data'] = column_data

        return ret

    def estimate_pairwise_row(self, tablename, function, row_list, clusters_name=None, threshold=None, modelids=None, numsamples=None):
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        if len(X_L_list) == 0:
            raise utils.BayesDBNoModelsError(tablename)

        # TODO: deal with row_list
        if row_list:
            row_indices = self.persistence_layer.get_row_list(tablename, row_list)
            if len(row_indices) == 0:
                raise utils.BayesDBError("Error: Row list %s has no rows." % row_list)
        else:
            row_indices = None

        column_lists = self.persistence_layer.get_column_lists(tablename)

        # Do the heavy lifting: generate the matrix itself

        matrix, row_indices_reordered, clusters = pairwise.generate_pairwise_row_matrix(function, X_L_list, X_D_list, M_c, T, tablename, engine=self, row_indices=row_indices, cluster_threshold=threshold, column_lists=column_lists, numsamples=numsamples)
        title = 'Pairwise row %s for %s' % (function.function_id, tablename)
        ret = dict(
            matrix=matrix,
            column_names=row_indices_reordered, # this is called column_names so that the plotting code displays them
            title=title,
            message = "Created " + title
            )

        # Create new btables from connected components (like into), if desired. Overwrites old ones with same name.
        if clusters is not None:
            cluster_name_tuples = []
            for i, cluster in enumerate(clusters):
                name = "%s_%d" % (clusters_name, i)
                num_rows = len(cluster)
                self.persistence_layer.add_row_list(tablename, name, cluster)
                cluster_name_tuples.append((name, num_rows))
            ret['clusters'] = clusters
            ret['row_lists'] = cluster_name_tuples

        return ret


    def estimate_pairwise(self, tablename, function_name, column_list=None, clusters_name=None, threshold=None, modelids=None, numsamples=None):
        if not self.persistence_layer.check_if_table_exists(tablename):
            raise utils.BayesDBInvalidBtableError(tablename)
        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        if len(X_L_list) == 0:
            raise utils.BayesDBNoModelsError(tablename)

        if column_list:
            column_names = self.persistence_layer.get_column_list(tablename, column_list)
            if len(column_names) == 0:
                raise utils.BayesDBError("Error: Column list %s has no columns." % column_list)

            utils.check_for_duplicate_columns(column_names)
        else:
            column_names = None

        # Do the heavy lifting: generate the matrix itself
        matrix, column_names_reordered, clusters = pairwise.generate_pairwise_column_matrix(
                function_name, X_L_list, X_D_list, M_c, T, tablename, engine=self, column_names=column_names,
                cluster_threshold=threshold, numsamples=numsamples)

        title = 'Pairwise column %s for %s' % (function_name, tablename)
        ret = dict(
            matrix=matrix,
            column_names=column_names_reordered,
            title=title,
            message = "Created " + title
            )

        # Add the column lists for connected clusters, if desired. Overwrites old ones with same name.
        if clusters is not None:
            cluster_name_tuples = []
            for i, cluster in enumerate(clusters):
                name = "%s_%d" % (clusters_name, i)
                column_names = [M_c['idx_to_name'][str(idx)] for idx in cluster]
                self.persistence_layer.add_column_list(tablename, name, column_names)
                cluster_name_tuples.append((name, column_names))
            ret['clusters'] = clusters
            ret['column_lists'] = cluster_name_tuples

        return ret

    def drop_column_list(self, tablename, list_name):
        """
        Remove column list from stored items for btable.
        """
        self.persistence_layer.drop_column_list(tablename, list_name)
        message = 'Column list %s removed from btable %s' % (list_name, tablename)
        return dict(message=message)

    def drop_row_list(self, tablename, list_name):
        """
        Remove row list from stored items for btable.
        """
        self.persistence_layer.drop_row_list(tablename, list_name)
        message = 'Row list %s removed from btable %s' % (list_name, tablename)
        return dict(message=message)

    def estimate_columns(self, tablename, functions, whereclause, limit, order_by, name=None, modelids=None, numsamples=None):
        """
        Return all the column names from the specified table as a list.
        First, columns are filtered based on whether they match the whereclause.
        The whereclause must consist of functions of a single column only.
        Next, the columns are ordered by other functions of a single column.
        Finally, the columns are limited to the specified number.

        ## allowed functions:
        # typicality(centrality)
        # dependence probability to <col>
        # mutual information with <col>
        # correlation with <col>
        """
        if not self.persistence_layer.check_if_table_exists(tablename):
          raise utils.BayesDBInvalidBtableError(tablename)

        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        column_indices = list(M_c['name_to_idx'].values())

        # Store the names of each of the functions in the where or order by, to be displayed in output.
        func_descriptions = []

        ## filter based on where clause
        where_conditions = self.parser.parse_column_whereclause(whereclause, M_c, T)
        if where_conditions is not None and len(X_L_list) == 0:
          raise utils.BayesDBNoModelsError(tablename)
        column_indices_and_data = estimate_columns_utils.filter_column_indices(column_indices, where_conditions, M_c, T, X_L_list, X_D_list, self, numsamples)
        func_descriptions += [estimate_columns_utils.function_description(where_item[0][0], where_item[0][1], M_c) for where_item in where_conditions]

        ## order
        if order_by and len(X_L_list) == 0:
          raise utils.BayesDBNoModelsError(tablename)
        if order_by != False:
          order_by = self.parser.parse_column_order_by_clause(order_by, M_c)
          func_descriptions += [estimate_columns_utils.function_description(order_item[0], order_item[1], M_c) for order_item in order_by]

        # Get tuples of column estimation function values and column indices
        ordered_tuples = estimate_columns_utils.order_columns(column_indices_and_data, order_by, M_c, X_L_list, X_D_list, T, self, numsamples)
        # tuple contains: (tuple(data), cidx)

        # limit
        if limit != float('inf'):
          ordered_tuples = ordered_tuples[:limit]

        # convert indices to names and create data list of column names and associated function values
        column_names = []
        column_data = []

        # column_data should be a list of the data values
        for ordered_tuple in ordered_tuples:
          if type(ordered_tuple) == int:
            #
            ordered_tuple = ((), column_idx_tup)
          column_name_temp = M_c['idx_to_name'][str(ordered_tuple[1])]
          column_names.append(column_name_temp)

          # Get the values of the functions that were used in where and order by
          column_data_temp = [data_utils.get_column_labels_from_M_c(M_c, [column_name_temp])[0]]
          column_data_temp.append(column_name_temp)
          column_data_temp.extend(ordered_tuple[0])
          column_data.append(column_data_temp)

        # save column list, if given a name to save as
        if name:
          # If list already exists with this name, warn and ask user to drop it.
          if self.persistence_layer.column_list_exists(tablename, name):
            print 'WARNING: Column list name %s already exists for btable %s and will not be overwritten.' % (name, tablename)
            print '    Use "DROP COLUMN LIST %s FROM %s" first.' % (name, tablename)
          else:
            self.persistence_layer.add_column_list(tablename, name, column_names)

        # De-duplicate values, if the same function was used in where and order by
        used_names = set()
        indices_to_pop = []
        func_descriptions.insert(0, 'column label')
        func_descriptions.insert(1, 'column name')
        for i, name in enumerate(func_descriptions):
          if name not in used_names:
            used_names.add(name)
          else:
            indices_to_pop.append(i)
        indices_to_pop.sort(reverse=True) # put highest indices first, so popping them doesn't affect others
        for i in indices_to_pop:
          func_descriptions.pop(i)
        for column_data_list in column_data:
          for i in indices_to_pop:
            column_data_list.pop(i)


        # Create column names ('column' followed by a string describing each function)
        ret = dict()
        if column_data != []:
          ret['column_labels'] = func_descriptions
          ret['data'] = column_data

        return ret

    def estimate_pairwise_row(self, tablename, function, row_list, clusters_name=None, threshold=None, modelids=None, numsamples=None):
        if not self.persistence_layer.check_if_table_exists(tablename):
          raise utils.BayesDBInvalidBtableError(tablename)
        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        if len(X_L_list) == 0:
          raise utils.BayesDBNoModelsError(tablename)

        # TODO: deal with row_list
        if row_list:
          row_indices = self.persistence_layer.get_row_list(tablename, row_list)
          if len(row_indices) == 0:
            raise utils.BayesDBError("Error: Row list %s has no rows." % row_list)
        else:
          row_indices = None

        column_lists = self.persistence_layer.get_column_lists(tablename)

        # Do the heavy lifting: generate the matrix itself

        matrix, row_indices_reordered, clusters = pairwise.generate_pairwise_row_matrix(function, X_L_list, X_D_list, M_c, T, tablename, engine=self, row_indices=row_indices, cluster_threshold=threshold, column_lists=column_lists, numsamples=numsamples)
        title = 'Pairwise row %s for %s' % (function.function_id, tablename)
        ret = dict(
          matrix=matrix,
          column_names=row_indices_reordered, # this is called column_names so that the plotting code displays them
          title=title,
          message = "Created " + title
          )

        # Create new btables from connected components (like into), if desired. Overwrites old ones with same name.
        if clusters is not None:
          # If lists with this prefix already exist, warn and ask user to remove them.
          if self.persistence_layer.row_list_exists(tablename, clusters_name):
            print 'WARNING: Column lists with prefix %s already exist for btable %s' % (clusters_name, tablename)
            print '    Use "DROP ROW LIST %s FROM %s" first.' % (clusters_name, tablename)
          else:
            cluster_name_tuples = []
            for i, cluster in enumerate(clusters):
              name = "%s_%d" % (clusters_name, i)
              num_rows = len(cluster)
              self.persistence_layer.add_row_list(tablename, name, cluster)
              cluster_name_tuples.append((name, num_rows))
            ret['clusters'] = clusters
            ret['row_lists'] = cluster_name_tuples

        return ret


    def estimate_pairwise(self, tablename, function_name, column_list=None, clusters_name=None, threshold=None, modelids=None, numsamples=None):
        if not self.persistence_layer.check_if_table_exists(tablename):
          raise utils.BayesDBInvalidBtableError(tablename)
        X_L_list, X_D_list, M_c = self.persistence_layer.get_latent_states(tablename, modelids)
        M_c, M_r, T = self.persistence_layer.get_metadata_and_table(tablename)
        if len(X_L_list) == 0:
          raise utils.BayesDBNoModelsError(tablename)

        if column_list:
          column_names = self.persistence_layer.get_column_list(tablename, column_list)
          if len(column_names) == 0:
            raise utils.BayesDBError("Error: Column list %s has no columns." % column_list)

          utils.check_for_duplicate_columns(column_names)
        else:
          column_names = None

        # Do the heavy lifting: generate the matrix itself
        matrix, column_names_reordered, clusters = pairwise.generate_pairwise_column_matrix(   \
            function_name, X_L_list, X_D_list, M_c, T, tablename, engine=self, column_names=column_names,
            cluster_threshold=threshold, numsamples=numsamples)

        title = 'Pairwise column %s for %s' % (function_name, tablename)
        ret = dict(
          matrix=matrix,
          column_names=column_names_reordered,
          title=title,
          message = "Created " + title
          )

        # Add the column lists for connected clusters, if desired. Overwrites old ones with same name.
        if clusters is not None:
          # If lists with this prefix already exist, warn and ask user to remove them.
          if self.persistence_layer.column_list_exists(tablename, clusters_name):
            print 'WARNING: Column lists with prefix %s already exist for btable %s' % (clusters_name, tablename)
            print '    Use "DROP COLUMN LIST %s FROM %s" first.' % (clusters_name, tablename)
          else:
            cluster_name_tuples = []
            for i, cluster in enumerate(clusters):
              name = "%s_%d" % (clusters_name, i)
              column_names = [M_c['idx_to_name'][str(idx)] for idx in cluster]
              self.persistence_layer.add_column_list(tablename, name, column_names)
              cluster_name_tuples.append((name, column_names))
            ret['clusters'] = clusters
            ret['column_lists'] = cluster_name_tuples

        return ret

# helper functions
get_name = lambda x: getattr(x, '__name__')
get_Engine_attr = lambda x: getattr(Engine, x)
is_Engine_method_name = lambda x: inspect.ismethod(get_Engine_attr(x))

def get_method_names():
        return filter(is_Engine_method_name, dir(Engine))

def get_method_name_to_args():
        method_names = get_method_names()
        method_name_to_args = dict()
        for method_name in method_names:
                method = Engine.__dict__[method_name]
                arg_str_list = inspect.getargspec(method).args[1:]
                method_name_to_args[method_name] = arg_str_list
        return method_name_to_args


class StoppableThread(threading.Thread):
        """Thread class with a stop() method. The thread itself has to check
        regularly for the stopped() condition."""

        def __init__(self, target, args):
                super(StoppableThread, self).__init__(target=target, args=args)
                self._stop = threading.Event()

        def stop(self):
                self._stop.set()

        def stopped(self):
                return self._stop.isSet()

class AnalyzeMaster(StoppableThread):
    def __init__(self, args):
        self.target = self.analyze_master
        super(AnalyzeMaster, self).__init__(target=self.target, args=args)

    def get_info(self):
        if not hasattr(self, 'start_time'):
            return "Analyze hasn't started"
        elapsed_seconds = time.time() - self.start_time
        if self.requested_iterations is None:
            # User specified time constraint
            remaining_seconds = self.total_seconds - elapsed_seconds
            if remaining_seconds > 0:
                return "Iterations completed: %d\nElapsed minutes: %0.2f\nRemaining minutes: %0.2f" % (self.iters_done, elapsed_seconds/60.0, remaining_seconds/60.0)
            else:
                return "Iterations completed: %d\nTime complete: waiting for final iteration to finish." % self.iters_done
        else:
            # User specified iterations constraint
            return "Iterations completed: %d\nRemaining iterations: %d\nElapsed minutes: %02.f\nEstimated remaining minutes:%0.2f" % (self.iters_done, self.requested_iterations - self.iters_done, elapsed_seconds/60.0, self.time_remaining_estimate/60.0)
        return timing_str

    def analyze_master(self, tablename, modelids, kernel_list, iterations, seconds, M_c, T, T_sub, models, background, engine):
        """
        Helper function for analyze. This is the thread that runs in the background as the previous analyze
        command returns before analyze is complete.
        """

        """ New stuff: scrutinize carefully
        analyze_args = dict(M_c=M_c, T=T, X_L=X_L_list, X_D=X_D_list, do_diagnostics=True,
                                                kernel_list=kernel_list)
        if ct_kernel != 0:
            analyze_args['CT_KERNEL'] = ct_kernel

        analyze_args['n_steps'] = iterations
        if seconds is not None:
            analyze_args['max_time'] = seconds
        """
        self.iters_done = 0
        self.time_remaining_estimate = -1
        # Even if limited by seconds, still limit to 1000 iterations.
        if iterations is None:
            self.requested_iterations = None
            iterations = 1000
        else:
            self.requested_iterations = iterations

        # Tracking info
        self.start_time = time.time()
        self.total_seconds = seconds

        ## CrossCat analyze works by trying to do the specified number of iterations (which MUST be at least 1,
        ##     and it never does more than that number), and doesn't start a new iteration if the specified time
        ##     has passed (although stopped after X seconds appeared to be buggy).

        ## If we have only a number of iterations: we'll do that many iterations
        ## If we have only a number of seconds: we'll keep starting new transitions up until that many seconds
        ##     has elapsed.
        ## If we have both: we'll do the min of the two.

        ## We will call CrossCat in 1 iteration batches for now, and in a future commit, we'll support
        ## batches of larger size for smaller tables.
        # TODO: ignores iterations

        # TODO: kill final model if over time? not important bc just running in background.

        X_L_list = [models[i]['X_L'] for i in models.keys()]
        X_D_list = [models[i]['X_D'] for i in models.keys()]

        threads = []
        # Create a thread for each modelid. Each thread will execute one iteration of analyze.
        for i, modelid in enumerate(modelids):
            if T_sub is None:
                p = AnalyzeWorker(args=(modelid, tablename, kernel_list, iterations, seconds, M_c, T, X_L_list[i], X_D_list[i], background, engine))
            else:
                p = AnalyzeWorker(args=(modelid, tablename, kernel_list, iterations, seconds, M_c, T_sub, X_L_list[i], X_D_list[i], background, engine))
            p.daemon = True
            p.start()
            threads.append(p)

        # before returning, make sure all threads have finished
        workers_done = False
        while not workers_done:
            time.sleep(.25)
            workers_done = True
            for p in threads:
                if not p.isAlive():
                    p.join()
                else:
                    workers_done = False

            # Get progress info
            info = zip(*[p.get_progress() for p in threads])
            iter_list = info[0]
            self.iters_done = min(iter_list)
            time_estimates = info[1]
            self.time_remaining_estimate = max(time_estimates)

            # If we are stopped, then stop all workers
            if self.stopped():
                for p in threads:
                    if p.isAlive():
                        p.stop()

        if T_sub is not None:
            engine._insert_subsampled_rows(tablename, T, T_sub, M_c, X_L_list, X_D_list, modelids)
        if background:
            print "\nAnalyze for table %s complete. Use 'SHOW MODELS FOR %s' to view models." % (tablename, tablename)

class AnalyzeWorker(StoppableThread):
    """ Controls ANALYZE for a single model.

    Attributes:
        target (function): analyze method
        iterations_done (int): Number of iterations performed by this worker
        time_per_model (float): Average time per iteration. Includes analyze and IO.
        collect_diagnostics_steps (int): Maximum number of steps between diagnostic data
            collections.
        collect_diagnostics_time (float): Maximum time (seconds) between diagnostic data
            collections.
        collection_ticker (int): Keeps track of the number of iterations since the last diagnostics
            collection.
        collection_timer (float): Keeps track for the time (seconds) since the last diagnostics
            collection.

    """
    def __init__(self, args):
        self.target = self.do_analyze_for_model
        super(AnalyzeWorker, self).__init__(target=self.target, args=args)
        self.iterations_done = 0
        self.time_per_model = 0
        self.collect_diagnostics_steps = 5
        self.collect_diagnostics_time = 5*60.0
        self.collection_ticker = 0
        self.collection_timer = 0

    def get_diagnostics(self, X_L, X_D):
        """Computes each single-state diagnostic available
        """
        diagnostics = dict()
        for item in diag_utils.single_state_diagnostics:
            item_method = getattr(diag_utils, 'get_'+item)
            diagnostics[item] = item_method(X_L, X_D)

        return diagnostics

    def get_progress(self):
        time_left = (self.iterations - self.iterations_done)*self.time_per_model
        return self.iterations_done, time_left

    def do_analyze_for_model(self, modelid, tablename, kernel_list, iterations, seconds, M_c, T,
                             X_L, X_D, background, engine):

        # TODO: this process should make a background thread for model writes!
        analyze_args = dict(M_c=M_c, T=T, do_diagnostics=True, kernel_list=kernel_list,
                            X_L=X_L, X_D=X_D, n_steps=1)

        start_time = time.time()
        last_write_time = start_time
        models_per_call = 1

        self.total_analyze_time = 0
        self.total_write_time = 0

        self.iterations = iterations
        self.iterations_done = 0

        self.collection_timer = 0.0
        self.collection_ticker = 0

        # diagnostics will be added to the model on each iteration.
        diagnostics_data = self.get_diagnostics(X_L, X_D)

        for i in range(iterations):
            cur_time = time.time()
            X_L, X_D, diagnostics_dict = engine.call_backend('analyze', analyze_args)

            this_iter_analyze_time = time.time() - cur_time
            self.collection_timer += this_iter_analyze_time
            self.collection_ticker += 1

            if self.collection_ticker == self.collect_diagnostics_steps or self.collection_timer > self.collect_diagnostics_time:
                diagnostics_data = self.get_diagnostics(X_L, X_D)
                diagnostics_data['logscore'] = diagnostics_dict['logscore'][-1]
                self.collection_timer = 0.0
                self.collection_ticker = 0

            analyze_args['X_L'] = X_L
            analyze_args['X_D'] = X_D

            cur_time = time.time()
            elapsed = cur_time - start_time
            engine.persistence_layer.update_model(tablename, X_L, X_D, diagnostics_data, modelid,
                                                  increment_iterations=1,
                                                  increment_time=this_iter_analyze_time)
            self.total_write_time += time.time() - cur_time
            self.iterations_done += 1

            if self.stopped() or (elapsed >= seconds and seconds is not None):
                return

            # Update number of models per call, with a target time per call of 5 seconds.
            self.time_per_model = float(elapsed)/self.iterations_done

            if self.time_per_model > 5:
                models_per_call = 1
            else:
                models_per_call = int(5.0/self.time_per_model)

            if iterations - self.iterations_done < models_per_call:
                models_per_call = iterations - self.iterations_done

            analyze_args['n_steps'] = models_per_call



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import sys
import csv
import copy
import pandas
import re
import numpy
import json
import prettytable
from math import pi

import utils


def get_ith_ordering(in_list, i):
    temp_list = [in_list[j::(i+1)][:] for j in range(i+1)]
    return [el for sub_list in temp_list for el in sub_list]


def gen_data(gen_seed, num_clusters, num_cols, num_rows, max_mean_per_category=10, max_std=1,
             max_mean=None):
    if max_mean is None:
        max_mean = max_mean_per_category * num_clusters
    n_grid = 11
    mu_grid = numpy.linspace(-max_mean, max_mean, n_grid)
    sigma_grid = 10 ** numpy.linspace(-1, numpy.log10(max_std), n_grid)
    num_rows_per_cluster = num_rows / num_clusters
    zs = numpy.repeat(range(num_clusters), num_rows_per_cluster)
    #
    random_state = numpy.random.RandomState(gen_seed)
    #
    data_size = (num_clusters, num_cols)
    which_mus = random_state.randint(len(mu_grid), size=data_size)
    which_sigmas = random_state.randint(len(sigma_grid), size=data_size)
    mus = mu_grid[which_mus]
    sigmas = sigma_grid[which_sigmas]
    clusters = []
    for row_mus, row_sigmas in zip(mus, sigmas):
        cluster_columns = []
        for mu, sigma in zip(row_mus, row_sigmas):
            cluster_column = random_state.normal(mu, sigma,
                                                 num_rows_per_cluster)
            cluster_columns.append(cluster_column)
        cluster = numpy.vstack(cluster_columns).T
        clusters.append(cluster)
    xs = numpy.vstack(clusters)
    return xs, zs


def gen_factorial_data(gen_seed, num_clusters, num_cols, num_rows, num_splits,
                       max_mean_per_category=10, max_std=1, max_mean=None):
    random_state = numpy.random.RandomState(gen_seed)
    data_list = []
    inverse_permutation_indices_list = []
    for data_idx in xrange(num_splits):
        data_i, zs_i = gen_data(
            gen_seed=random_state.randint(sys.maxint),
            num_clusters=num_clusters,
            num_cols=num_cols/num_splits,
            num_rows=num_rows,
            max_mean_per_category=max_mean_per_category,
            max_std=max_std,
            max_mean=max_mean
            )
        permutation_indices = random_state.permutation(xrange(num_rows))
        # permutation_indices = get_ith_ordering(range(num_rows), data_idx)
        inverse_permutation_indices = numpy.argsort(permutation_indices)
        inverse_permutation_indices_list.append(inverse_permutation_indices)
        data_list.append(numpy.array(data_i)[permutation_indices])
    data = numpy.hstack(data_list)
    return data, inverse_permutation_indices_list


def gen_raw_T_full_from_T_full(T_full, M_c):
    """
    engine.upgrade_btables needs a way to go from T to raw_T_full for tables
    saved without a metadata_full file. This function takes T and M_c
    and reconstructs the raw data list.
    """
    raw_T_full = T_full
    for row_idx, row_data in enumerate(T_full):
        for col_idx, code in enumerate(row_data):
            code = T_full[row_idx][col_idx]
            raw_T_full[row_idx][col_idx] = str(convert_code_to_value(M_c, col_idx, code))
    return raw_T_full


def gen_M_r_from_T(T):
    num_rows = len(T)
    num_cols = len(T[0])
    #
    name_to_idx = dict(zip(map(str, range(num_rows)), range(num_rows)))
    idx_to_name = dict(zip(map(str, range(num_rows)), range(num_rows)))
    M_r = dict(name_to_idx=name_to_idx, idx_to_name=idx_to_name)
    return M_r


def gen_ignore_metadata(column_data, parameters=None):
    ret = gen_categorical_metadata(column_data, parameters)
    ret['modeltype'] = 'ignore'
    ret['parameters'] = None
    return ret


def gen_numerical_metadata(column_data, parameters=None):
    res = dict(
        modeltype="normal_inverse_gamma",
        value_to_code=dict(),
        code_to_value=dict(),
        parameters=None
    )
    return res


def gen_cyclic_metadata(column_data, parameters=None):

    data_min = min(map(float, column_data))
    data_max = max(map(float, column_data))

    if not parameters:
        parameters = dict(min=data_min, max=data_max)
    else:
        if 'min' not in parameters or 'max' not in parameters:
            raise utils.BayesDBError("Error: cyclic columns require (min, max) parameters."
                                     % str(value))
        else:
            param_min = float(parameters['min'])
            param_max = float(parameters['max'])
            if data_min < param_min:
                raise utils.BayesDBError("Error: cyclic contains data less than specified "
                                         "minimum %f" % param_min)
            elif data_max > param_max:
                raise utils.BayesDBError("Error: cyclic contains data greater than specified "
                                         "maximum %f" % param_max)
            else:
                parameters = dict(min=param_min, max=param_max)

    return dict(
        modeltype="vonmises",
        value_to_code=dict(),
        code_to_value=dict(),
        parameters=parameters
        )


def gen_categorical_metadata(column_data, parameters=None):
    def get_is_not_nan(el):
        if isinstance(el, str):
            return el.upper() != 'NAN'
        else:
            return True
    # get_is_not_nan = lambda el: el.upper() != 'NAN'
    #
    unique_codes = list(set(column_data))
    unique_codes = filter(get_is_not_nan, unique_codes)
    #
    values = range(len(unique_codes))
    value_to_code = dict(zip(values, unique_codes))
    code_to_value = dict(zip(unique_codes, values))

    # Set cardinality = number of distinct values if not set, otherwise check
    # that cardinality parameter is >= the number of distinct values.
    n_codes = len(unique_codes)
    if not parameters:
        parameters = dict(cardinality=n_codes)
    else:
        parameters['cardinality'] = int(parameters['cardinality'])
        if n_codes > parameters['cardinality']:
            raise utils.BayesDBError("Error: categorical contains more distinct values than "
                                     "specified cardinality %i" % parameters['cardinality'])

    ret = dict(
        modeltype="symmetric_dirichlet_discrete",
        value_to_code=value_to_code,
        code_to_value=code_to_value,
        parameters=parameters
    )

    return ret

metadata_generator_lookup = dict(
    numerical=gen_numerical_metadata,
    cyclic=gen_cyclic_metadata,
    categorical=gen_categorical_metadata,
    ignore=gen_ignore_metadata,
    key=gen_ignore_metadata,
)


def gen_M_c_from_T(T, cctypes=None, colnames=None, parameters=None, codebook=None):
    num_rows = len(T)
    num_cols = len(T[0])
    if cctypes is None:
        cctypes = ['numerical'] * num_cols
    if colnames is None:
        colnames = range(num_cols)
    if parameters is None:
        parameters = [None] * num_cols
    #
    T_array_transpose = numpy.array(T).T
    column_metadata = []
    for cctype, column_data, params in zip(cctypes, T_array_transpose, parameters):
        metadata_generator = metadata_generator_lookup[cctype]
        metadata = metadata_generator(column_data, params)
        column_metadata.append(metadata)

    column_codebook = []
    for col_idx, colname in enumerate(colnames):
        if codebook and colname in codebook:
            colname_codebook = codebook[colname]

            # update column metadata with value maps
            if colname_codebook['value_map'].upper() != 'NAN':
                if column_metadata[col_idx]['modeltype'] != 'symmetric_dirichlet_discrete':
                    raise utils.BayesDBError('Value map specified for non-categorical column({})').format(colname)

                try:
                    colvm = json.loads(colname_codebook['value_map'])
                    codes = colvm.keys()
                    values = colvm.values()
                    column_metadata[col_idx]['code_to_value'] = colvm
                    column_metadata[col_idx]['value_to_code'] = dict(zip(values, codes))
                    column_metadata[col_idx]['parameters'] = {'cardinality':  len(values)}
                except:
                    raise utils.BayesDBError('Error parsing vaue map in codebook for {}.'.format(colname))
        else:
            colname_codebook = {
                'description': 'No description',
                'short_name': colname,
                'value_map': None
            }
        column_codebook.append(colname_codebook)

    name_to_idx = dict(zip(colnames, range(num_cols)))
    idx_to_name = dict(zip(map(str, range(num_cols)), colnames))

    M_c = dict(
        name_to_idx=name_to_idx,
        idx_to_name=idx_to_name,
        column_metadata=column_metadata,
        column_codebook=column_codebook
        )
    return M_c


def gen_M_c_from_T_with_colnames(T, colnames):
    num_rows = len(T)
    num_cols = len(T[0])
    #
    gen_numerical_metadata = lambda: dict(modeltype="normal_inverse_gamma",
                                          value_to_code=dict(),
                                          code_to_value=dict())
    column_metadata = [
        gen_numerical_metadata()
        for col_idx in range(num_cols)
        ]
    name_to_idx = dict(zip(colnames, range(num_cols)))
    idx_to_name = dict(zip(map(str, range(num_cols)), colnames))
    M_c = dict(
        name_to_idx=name_to_idx,
        idx_to_name=idx_to_name,
        column_metadata=column_metadata,
        )
    return M_c


def gen_factorial_data_objects(gen_seed, num_clusters, num_cols, num_rows, num_splits, max_mean=10,
                               max_std=1, send_data_inverse_permutation_indices=False):
    T, data_inverse_permutation_indices = gen_factorial_data(gen_seed, num_clusters, num_cols,
                                                             num_rows, num_splits, max_mean,
                                                             max_std)
    T = T.tolist()
    M_r = gen_M_r_from_T(T)
    M_c = gen_M_c_from_T(T)

    if not send_data_inverse_permutation_indices:
        return T, M_r, M_c
    else:
        return T, M_r, M_c, data_inverse_permutation_indices


def discretize_data(T, discretize_indices):
    T_array = numpy.array(T)
    discretize_indices = numpy.array(discretize_indices)
    T_array[:, discretize_indices] = \
        numpy.array(T_array[:, discretize_indices], dtype=int)
    return T_array.tolist()


def convert_columns_to_categorical(T, M_c, categorical_indices):
    categorical_indices = numpy.array(categorical_indices)
    modeltype = 'symmetric_dirichlet_discrete'
    T_array = numpy.array(T)
    for categorical_idx in categorical_indices:
        categorical_column = T_array[:, categorical_idx]
        categorical_column = categorical_column[~numpy.isnan(categorical_column)]
        categorical_values = list(set(categorical_column))
        K = len(categorical_values)
        code_to_value = dict(zip(range(K), categorical_values))
        value_to_code = dict(zip(categorical_values, range(K)))
        categorical_column_metadata = M_c['column_metadata'][categorical_idx]
        categorical_column_metadata['modeltype'] = modeltype
        categorical_column_metadata['code_to_value'] = code_to_value
        categorical_column_metadata['value_to_code'] = value_to_code
    return T, M_c


# UNTESTED
def convert_columns_to_numerical(T, M_c, numerical_indices):
    numerical_indices = numpy.array(numerical_indices)
    modeltype = 'normal_inverse_gamma'
    T_array = numpy.array(T)
    for numerical_idx in numerical_indices:
        code_to_value = dict()
        value_to_code = dict()
        numerical_column_metadata = M_c['column_metadata'][numerical_idx]
        numerical_column_metadata['modeltype'] = modeltype
        numerical_column_metadata['code_to_value'] = code_to_value
        numerical_column_metadata['value_to_code'] = value_to_code
    return T, M_c


def at_most_N_rows(T, N, gen_seed=0):
    num_rows = len(T)
    if (N is not None) and (num_rows > N):
        random_state = numpy.random.RandomState(gen_seed)
        which_rows = random_state.permutation(xrange(num_rows))
        which_rows = which_rows[:N]
        T = [T[which_row] for which_row in which_rows]
    return T


def construct_pandas_df(query_obj):
    """
    Take a result from a BQL statement (dict with 'data' and 'columns')
    and constructs a pandas data frame.

    Currently this is only called if the user specifies pandas_output = True
    """
    if len(query_obj['data']) == 0:
        data = None
    else:
        # Some types (numpy recarray for one) caused some issues when read into
        # a DataFrame, so recasting this as a list of lists solves the problem
        data = [list(row) for row in query_obj['data']]

    # For data queries, we want column_names as the column names in the pandas output.
    # For non-data queries, we only have column_labels, so use that as backup.
    if 'column_names' in query_obj:
        columns = query_obj['column_names']
    else:
        columns = query_obj['column_labels']

    pandas_df = pandas.DataFrame(data=data, columns=columns)
    return pandas_df


def read_pandas_df(pandas_df):
    """
    Takes pandas data frame object and converts data
    into list-of-lists format
    """
    header = list(pandas_df.columns)
    rows = [map(str, row) for index, row in pandas_df.iterrows()]
    return header, rows


def read_csv(filename, has_header=True):
    with open(filename, 'rU') as fh:
        csv_reader = csv.reader(fh)
        header = None
        if has_header:
            header = csv_reader.next()
        rows = [[r.strip() for r in row] for row in csv_reader]
    return header, rows


def write_csv(filename, T, header=None):
    with open(filename, 'w') as fh:
        csv_writer = csv.writer(fh, delimiter=',')
        if header is None:
            csv_writer.writerow(header)
        [csv_writer.writerow(T[i]) for i in range(len(T))]


def all_numerical_from_file(filename, max_rows=None, gen_seed=0, has_header=True):
    header, T = read_csv(filename, has_header=has_header)
    T = numpy.array(T, dtype=float).tolist()
    T = at_most_N_rows(T, N=max_rows, gen_seed=gen_seed)
    M_r = gen_M_r_from_T(T)
    M_c = gen_M_c_from_T(T)
    return T, M_r, M_c, header


def numerical_or_ignore_from_file_with_colnames(filename, cctypes, max_rows=None, gen_seed=0):
    header = None
    T, M_r, M_c = None, None, None
    colmask = map(lambda x: 1 if x != 'ignore' else 0, cctypes)
    with open(filename) as fh:
        csv_reader = csv.reader(fh)
        header = csv_reader.next()
        # TODO: why is this list then an array and then a list again?
        T = numpy.array([[col for col, flag in zip(row, colmask) if flag] for row in csv_reader],
                        dtype=float).tolist()
        num_rows = len(T)
        if (max_rows is not None) and (num_rows > max_rows):
            random_state = numpy.random.RandomState(gen_seed)
            which_rows = random_state.permutation(xrange(num_rows))
            which_rows = which_rows[:max_rows]
            T = [T[which_row] for which_row in which_rows]
        M_r = gen_M_r_from_T(T)
        M_c = gen_M_c_from_T_with_colnames(T, [col for col, flag in zip(header, colmask) if flag])
    return T, M_r, M_c, header


def convert_code_to_value(M_c, cidx, code):
    """
    For a column with categorical data, this function takes the 'code':
    the integer used to represent a specific value, and returns the corresponding
    raw value (e.g. 'Joe' or 234.23409), which is always encoded as a string.

    Note that the underlying store 'value_to_code' is unfortunately named backwards.
    TODO: fix the backwards naming.
    """
    if flexible_isnan(code):
        return code

    column_metadata = M_c['column_metadata'][cidx]
    modeltype = column_metadata['modeltype']
    if modeltype == 'normal_inverse_gamma':
        return float(code)
    elif modeltype == 'vonmises':
        # Convert stored value (in [0, 2pi]) back to raw range.
        param_min = column_metadata['parameters']['min']
        param_max = column_metadata['parameters']['max']
        return param_min + ((float(code) / (2 * pi)) * (param_max - param_min))
    else:
        try:
            return M_c['column_metadata'][cidx]['value_to_code'][int(code)]
        except KeyError:
            return M_c['column_metadata'][cidx]['value_to_code'][str(int(code))]


def convert_value_to_code(M_c, cidx, value):
    """
    For a column with categorical data, this function takes the raw value
    (e.g. 'Joe' or 234.23409), which is always encoded as a string, and returns the
    'code': the integer used to represent that value in the underlying representation.

    Note that the underlying store 'code_to_value' is unfortunately named backwards.
    TODO: fix the backwards naming.
    """
    column_metadata = M_c['column_metadata'][cidx]
    modeltype = column_metadata['modeltype']
    if modeltype == 'normal_inverse_gamma':
        return float(value)
    elif modeltype == 'vonmises':
        param_min = column_metadata['parameters']['min']
        param_max = column_metadata['parameters']['max']
        return (float(value) - param_min) / (param_max - param_min)
    else:
        try:
            return M_c['column_metadata'][cidx]['code_to_value'][str(value)]
        except KeyError:
            raise utils.BayesDBError("Error: value '%s' not in btable." % str(value))


def map_from_T_with_M_c(coordinate_value_tuples, M_c):
    coordinate_code_tuples = []
    column_metadata = M_c['column_metadata']
    for row_idx, col_idx, value in coordinate_value_tuples:
        datatype = column_metadata[col_idx]['modeltype']
        # FIXME: make this robust to different datatypes
        if datatype == 'symmetric_dirichlet_discrete':
            # FIXME: casting key to str is a hack
            value = column_metadata[col_idx]['value_to_code'][str(int(value))]
        coordinate_code_tuples.append((row_idx, col_idx, value))
    return coordinate_code_tuples


def map_to_T_with_M_c(T_uncast_array, M_c):
    T_uncast_array = numpy.array(T_uncast_array)
    # WARNING: array argument is mutated
    for col_idx in range(T_uncast_array.shape[1]):
        column_metadata = M_c['column_metadata'][col_idx]
        modeltype = column_metadata['modeltype']
        col_data = T_uncast_array[:, col_idx]

        if modeltype == 'normal_inverse_gamma':
            continue
        elif modeltype == 'vonmises':
            param_min = column_metadata['parameters']['min']
            param_max = column_metadata['parameters']['max']
            mapped_values = [2 * pi * (float(x) - param_min) / (param_max - param_min)
                             for x in col_data]
        else:
            # copy.copy else you mutate M_c
            mapping = copy.copy(M_c['column_metadata'][col_idx]['code_to_value'])
            mapping['NAN'] = numpy.nan
            to_upper = lambda el: str(el).upper()
            is_nan_str = numpy.array(map(to_upper, col_data)) == 'NAN'
            col_data[is_nan_str] = 'NAN'
            # FIXME: THIS IS WHERE TO PUT NAN HANDLING
            mapped_values = [mapping[el] for el in col_data]
        T_uncast_array[:, col_idx] = mapped_values
    T = numpy.array(T_uncast_array, dtype=float).tolist()
    return T


def do_pop_list_indices(in_list, pop_indices):
    pop_indices = sorted(pop_indices, reverse=True)
    _do_pop = lambda x: in_list.pop(x)
    map(_do_pop, pop_indices)
    return in_list


def get_list_indices(in_list, get_indices_of):
    lookup = dict(zip(in_list, range(len(in_list))))
    indices = map(lookup.get, get_indices_of)
    # The line below removes [0] from indices. Need to check with Jay about how to handle those
    # cases.
    # indices = filter(None, indices)
    return indices


def transpose_list(in_list):
    return zip(*in_list)


def get_pop_indices(cctypes, colnames):
    assert len(colnames) == len(cctypes)
    pop_columns = [colname for (cctype, colname) in zip(cctypes, colnames)
                   if (cctype in ['ignore', 'key'])]
    pop_indices = get_list_indices(colnames, pop_columns)
    return pop_indices


def do_pop_columns(T, pop_indices):
    T_by_columns = transpose_list(T)
    T_by_columns = do_pop_list_indices(T_by_columns, pop_indices)
    T = transpose_list(T_by_columns)
    return T


def remove_ignore_cols(T, cctypes, colnames, parameters=None):
    pop_indices = get_pop_indices(cctypes, colnames)
    T = do_pop_columns(T, pop_indices)
    colnames = do_pop_list_indices(colnames[:], pop_indices)
    cctypes = do_pop_list_indices(cctypes[:], pop_indices)
    if parameters:
        parameters = do_pop_list_indices(parameters[:], pop_indices)
        ret = T, cctypes, colnames, parameters
    else:
        ret = T, cctypes, colnames
    return ret

nan_set = set(['', 'null', 'n/a'])
_convert_nan = lambda el: el if str(el).strip().lower() not in nan_set else 'NAN'
_convert_nans = lambda in_list: map(_convert_nan, in_list)
convert_nans = lambda in_T: map(_convert_nans, in_T)


def flexible_isnan(val):
    try:
        if val in nan_set or numpy.isnan(float(val)):
            return True
    except:
        pass
    return False


def read_data_objects(filename, max_rows=None, gen_seed=0,
                      cctypes=None, colnames=None):
    header, raw_T = read_csv(filename, has_header=True)
    header = [h.lower().strip() for h in header]
    # FIXME: why both accept colnames argument and read header?
    if colnames is None:
        colnames = header
        pass
    # remove excess rows
    raw_T = at_most_N_rows(raw_T, N=max_rows, gen_seed=gen_seed)
    raw_T = convert_nans(raw_T)

    if cctypes is None:
        cctypes = ['numerical'] * len(header)
        pass

    # remove ignore columns
    T_uncast_arr, cctypes, header = remove_ignore_cols(raw_T, cctypes, header)
    # determine value mappings and map T to numerical castable values
    M_r = gen_M_r_from_T(T_uncast_arr)
    M_c = gen_M_c_from_T(T_uncast_arr, cctypes, colnames)
    T = map_to_T_with_M_c(T_uncast_arr, M_c)
    #
    return T, M_r, M_c, header


def get_can_cast_to_float(column_data):
    can_cast = True
    try:
        [float(datum) for datum in column_data]
    except:
        can_cast = False
    return can_cast


def get_can_cast_to_int(column_data):
    can_cast = True
    try:
        [int(datum) for datum in column_data]
    except:
        can_cast = False
    return can_cast


def get_int_equals_str(column_data):
    try:
        equals = all([str(datum) == str(int(datum)) for datum in column_data])
    except:
        equals = False
    return equals


def guess_column_type(column_data, count_cutoff=20):
    num_distinct = len(set(column_data))
    above_count_cutoff = num_distinct > count_cutoff
    can_cast = get_can_cast_to_float(column_data)
    if above_count_cutoff and can_cast:
        column_type = 'numerical'
    else:
        column_type = 'categorical'
    return column_type


def guess_column_types(T, colnames_full, count_cutoff=20, warn_ratio=0.02, warn_cardinality=7):
    """
    Guesses column types - used when creating new btable so user doesn't have to
    specify a type for all columns.
    Refer to function guess_column_type for decision rules.
    Warn if cardinality of a categorical is greater than specified (default 7)
    or if distinctness ratio of numerical is lower than specified (default 0.02).
        (limit proposed by Pat Shafto 7 Aug 2014)
    """
    T_transposed = transpose_list(T)
    column_types = []
    warnings = []
    for column_idx, column_data in enumerate(T_transposed):
        column_type = guess_column_type(column_data, count_cutoff)
        column_types.append(column_type)
        num_data = len(column_data)
        num_distinct = len(set(column_data))
        distinct_ratio = float(num_distinct) / num_data
        if column_type == 'categorical':
            if num_distinct > warn_cardinality:
                warnings.append('Column "%s" is categorical but has a high number of distinct values. '
                            'Convert to numerical using UPDATE SCHEMA if appropriate.'
                            % colnames_full[column_idx])
        elif distinct_ratio < warn_ratio:
            warnings.append('Column "%s" is NOT categorical but has a low percentage of distinct values '
                            '( actual: %0.2f   <   warn threshold: %0.2f ). '
                            'Convert to categorical using UPDATE SCHEMA if appropriate.'
                            % (colnames_full[column_idx], distinct_ratio, warn_ratio))
    return column_types, warnings


def read_model_data_from_csv(filename, max_rows=None, gen_seed=0, cctypes=None):
    colnames, T = read_csv(filename)
    return gen_T_and_metadata(colnames, raw_T, max_rows, gen_seed, cctypes)


def gen_T_and_metadata(colnames, raw_T, max_rows=None, gen_seed=0,
                       cctypes=None, parameters=None, codebook=None):
    T = at_most_N_rows(raw_T, max_rows, gen_seed)
    T = convert_nans(T)
    if cctypes is None:
        cctypes = guess_column_types(T)
    M_c = gen_M_c_from_T(T, cctypes, colnames, parameters, codebook)
    T = map_to_T_with_M_c(numpy.array(T), M_c)
    M_r = gen_M_r_from_T(T)
    return T, M_r, M_c, cctypes

extract_view_count = lambda X_L: len(X_L['view_state'])
extract_cluster_count = lambda view_state_i: view_state_i['row_partition_model']['counts']
extract_cluster_counts = lambda X_L: map(extract_cluster_count, X_L['view_state'])
get_state_shape = lambda X_L: (extract_view_count(X_L), extract_cluster_counts(X_L))


def is_key_eligible(x):
    """
    A column is eligible to be the table key if:
    All values are unique, AND
    can't be cast to float (a string key) OR string representation of all values matches the
    string representation of its int value
    """
    values_unique = len(x) == len(set(x))
    castable = not get_can_cast_to_float(x) or get_int_equals_str(x)
    return values_unique and castable


def select_key_column(raw_T_full, colnames_full, cctypes_full, key_column=None, testing=False):
    """
    This function takes the raw data, colnames, and data types from an input CSV file from
    which a btable is being created.

    It looks for every column that's eligible to be a table key, and prompts the user to select one,
    or create a new key. If no eligible key is found, a new one is created by default.

    That way a btable can't be created without a key column.
    """
    T_df = pandas.DataFrame(data=raw_T_full, columns=colnames_full)
    eligibility = T_df.apply(is_key_eligible)

    key_eligibles = list(eligibility[eligibility].index)
    key_eligibles_len = len(key_eligibles)
    if testing:
        key_column_selection = 0
    elif key_eligibles_len == 0 and key_column is None:
        print("None of the columns in this table is eligible to be the key. A key column will be "
              "created. Press any key continue.")
        user_confirmation = raw_input()
        key_column_selection = 0
    elif key_column is None or key_column not in range(key_eligibles_len + 1):
        key_column_selection = None
        pt = prettytable.PrettyTable()
        pt.field_names = ['choice', 'key column']
        pt.add_row([0, '<Create key column>'])
        for index, key_eligible in enumerate(key_eligibles):
            pt.add_row([index + 1, key_eligible])
        while key_column_selection is None:
            print(str(pt))
            print("Please select which column you would like to set as the table key:")
            user_selection = raw_input()
            try:
                user_selection = int(user_selection)
                if user_selection in range(key_eligibles_len + 1):
                    key_column_selection = user_selection
            except:
                continue
    else:
        key_column_selection = key_column

    # 0 always means insert a new key column.
    if key_column_selection == 0:
        raw_T_full, colnames_full, cctypes_full, key_column = insert_key_column(T_df, colnames_full,
                                                                                cctypes_full)
    else:
        key_column = key_eligibles[key_column_selection - 1]
        cctypes_full[colnames_full.index(key_column)] = 'key'

    return raw_T_full, colnames_full, cctypes_full


def insert_key_column(T_df, colnames_full, cctypes_full):
    """
    Create a new table key column. It'll be the left-most column and an ascending integer column,
    with name 'key' and cctype 'key'.
    """
    key_column_name = 'key'
    if key_column_name in colnames_full:
        key_column_suffix = 0
        while key_column_name + str(key_column_suffix) in colnames_full:
            key_column_suffix += 1
        key_column_name = key_column_name + str(key_column_suffix)
    T_df.insert(0, key_column_name, map(str, range(T_df.shape[0])))
    raw_T_full = T_df.to_records(index=False)
    colnames_full.insert(0, key_column_name)
    cctypes_full.insert(0, 'key')
    return raw_T_full, colnames_full, cctypes_full, key_column_name


def get_column_labels_from_M_c(M_c, colnames):
    """
    Get display labels from M_c by pulling from codebook.
    Sources in order of priority:
        1. column_codebook[col_idx][short_name]
        2. colname
    """

    # Initialize as colnames
    column_labels = colnames[:]

    for colname_idx, colname in enumerate(colnames):
        if colname in M_c['name_to_idx']:
            colname_idx_M_c = M_c['name_to_idx'][colname]
            column_codebook = M_c['column_codebook'][colname_idx_M_c]

            if column_codebook:
                short_name = column_codebook['short_name']
                if short_name != '':
                    column_labels[colname_idx] = short_name

    return column_labels



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from pyparsing import Literal, CaselessLiteral, Word, Upcase, delimitedList, Optional,\
    Combine, Group, alphas, nums, alphanums, ParseException, Forward, oneOf, quotedString,\
    ZeroOrMore, OneOrMore, restOfLine, Keyword

selectStatement = Forward()

identifier = Word(alphas, alphanums + "_")


selectToken = Keyword("select", caseless=True)
fromToken = Keyword("from", caseless=True)
orderByToken = Keyword("order", caseless=True) + Keyword("by", caseless=True)
limitToken = Keyword("limit", caseless=True)

columnNameList = Group(delimitedList(identifier | '*'))

createBtableStatement = Keyword("create", caseless=True) + Keyword("btable", caseless=True) + \
                        identifier.setResultsName("tablename") + fromToken + identifier.setResultsName("filename")

#orderByClause = orderByToken + 

selectStatement << ( selectToken +
                     columnNameList.setResultsName("columns") +
                     fromToken +
                     identifier.setResultsName("tablename") +
                     Optional(whereClause) +
                     Optional(orderByClause) +
                     Optional(limitClause) )

BQLStatement = (selectStatement | createBtableStatement) + Optional(';')

BQL = ZeroOrMore(BQLStatement)

## allows comments
dashComment = "--" + restOfLine
BQL.ignore(dashComment)



def test( str ):
    print str,"->"
    try:
        tokens = BQL.parseString( str )
        print "tokens = ",        tokens
        print "tokens.tablename =", tokens.tablename
        print "tokens.filename =",  tokens.filename
        #print "tokens.where =", tokens.where
    except ParseException, err:
        print " "*err.loc + "^\n" + err.msg
        print err
    print


class PyParser(object):
    def __init__(self):
        pass

    def parse(self, bql_string):
        pass

    def parse_line(self, bql_string):
        pass



#
#   Copyright (c) 2010-2014, MIT Probabilistic Computing Project
#
#   Lead Developers: Jay Baxter and Dan Lovell
#   Authors: Jay Baxter, Dan Lovell, Baxter Eaves, Vikash Mansinghka
#   Research Leads: Vikash Mansinghka, Patrick Shafto
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import time
import inspect
import pickle
import os
import numpy
import pytest
import random

import bayesdb.diagnostics_utils as diag_utils
import bayesdb.data_utils as data_utils
from bayesdb.client import Client
from bayesdb.engine import Engine
from bayesdb.parser import Parser
import bayesdb.bql_grammar as bql
import bayesdb.utils as utils

engine = Engine()
parser = Parser()

test_tablenames = None


notimplemented = pytest.mark.skipif(True, reason="Test not implemented")


def setup_function(function):
    global test_tablenames
    test_tablenames = []
    global engine
    engine = Engine()


def teardown_function(function):
    global tablename
    for test_tablename in test_tablenames:
        engine.drop_btable(test_tablename)


def create_dha(path='data/dha.csv'):
    test_tablename = 'dhatest' + str(int(time.time() * 1000000)) + \
        str(int(random.random()*10000000))
    header, rows = data_utils.read_csv(path)
    create_btable_result = engine.create_btable(test_tablename, header, rows, key_column=0)
    # metadata = engine.persistence_layer.get_metadata(test_tablename)

    global test_tablenames
    test_tablenames.append(test_tablename)

    return test_tablename, create_btable_result


def create_describe_btable(data_path='data/describe.csv',
                           codebook_path='data/describe_codebook.csv', use_codebook=True):
    # TODO: refactor codebook generation to Engine, not Client
    test_tablename = 'describetest' + str(int(time.time() * 1000000)) + \
        str(int(random.random()*10000000))

    if use_codebook:
        codebook_header, codebook_rows = data_utils.read_csv(codebook_path)
        codebook = dict()
        for codebook_row in codebook_rows:
            codebook[codebook_row[0]] = dict(zip(['short_name', 'description', 'value_map'],
                                             codebook_row[1:]))
    else:
        codebook = None

    header, rows = data_utils.read_csv(data_path)
    create_btable_result = engine.create_btable(test_tablename, header, rows, key_column=0,
                                                codebook=codebook)

    global test_tablenames
    test_tablenames.append(test_tablename)

    return test_tablename, create_btable_result


def test_create_btable():
    test_tablename, create_btable_result = create_dha()
    assert 'column_labels' in create_btable_result
    assert 'data' in create_btable_result
    assert 'message' in create_btable_result
    # Should be 65 rows (1 for each column inferred: 64 from data file, plus 1 for key)
    assert len(create_btable_result['data']) == 65
    # Should be 3 columns: 1 for column names, 1 for data type, and 1 for parameters
    assert len(create_btable_result['data'][0]) == 3
    list_btables_result = engine.list_btables()['data']
    assert [test_tablename] in list_btables_result
    engine.drop_btable(test_tablename)


def test_drop_btable():
    test_tablename, _ = create_dha()
    list_btables_result = engine.list_btables()['data']
    assert [test_tablename] in list_btables_result
    engine.drop_btable(test_tablename)
    list_btables_result = engine.list_btables()['data']
    assert [test_tablename] not in list_btables_result


def test_select():
    test_tablename, _ = create_dha()

    # Test a simple query: select two columns, no limit, no order, no where.
    # Check to make sure types of all inputs are correct, etc.

    functions = bql.bql_statement.parseString('select name, qual_score from test',
                                              parseAll=True).functions
    whereclause = None
    limit = float('inf')
    order_by = False
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert 'column_labels' in select_result
    assert 'data' in select_result
    assert select_result['column_labels'] == ['key', 'name', 'qual_score']
    # 307 is the total number of rows in the dataset.
    assert(len(select_result['data']) == 307 and
           len(select_result['data'][0]) == len(select_result['column_labels']))
    assert type(select_result['data'][0][0]) == numpy.string_
    t = type(select_result['data'][0][1])
    assert (t == unicode) or (t == str) or (t == numpy.string_)  # type of name is unicode or string
    assert type(select_result['data'][0][2]) == float  # type of qual_score is float
    original_select_result = select_result['data']

    # Test limit: do the same query as before, but limit to 10
    limit = 10
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert len(select_result['data']) == limit

    # Test order by single column: desc
    ground_truth_ordered_results = sorted(original_select_result, key=lambda t: t[2],
                                          reverse=True)[:10]
    order_by = [('qual_score', True)]
    order_by = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                             'order by qual_score desc', parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert select_result['data'] == ground_truth_ordered_results

    # Test order by single column: asc
    ground_truth_ordered_results = sorted(original_select_result, key=lambda t: t[2])[:10]
    order_by = [('qual_score', False)]
    order_by = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                             'order by qual_score asc', parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert select_result['data'] == ground_truth_ordered_results

    engine.initialize_models(test_tablename, 2)

    # SIMILARITY TO <row> [WITH RESPECT TO <col>]
    # smoke tests
    functions = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                              'order by similarity to 5', parseAll=True).functions
    order_by = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                             'order by similarity to 5', parseAll=True).order_by

    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                              'order by similarity to 5', parseAll=True).functions
    order_by = bql.bql_statement.parseString('select name, qual_score, similarity to 5 from test '
                                             'order by similarity to 5 with respect to qual_score',
                                             parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString('select name, qual_score from test',
                                              parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by similarity to 5 with '
                                             'respect to qual_score', parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString('select name, qual_score, similarity to 5 with '
                                              'respect to name from test', parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by similarity to 5',
                                             parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString("select name, qual_score, similarity to "
                                              "name='Albany NY' with respect to qual_score from "
                                              "test", parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by similarity to 5',
                                             parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString('select * from test', parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by similarity to 5 with '
                                             'respect to name', parseAll=True).order_by
    whereclause = bql.bql_statement.parseString('select * from test where qual_score > 6',
                                                parseAll=True).where_conditions
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    # Albany NY's row id is 3
    whereclause = bql.bql_statement.parseString('select * from test where name="Albany NY"',
                                                parseAll=True).where_conditions
    functions = bql.bql_statement.parseString('select * from test', parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by similarity to 5 with '
                                             'respect to name', parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    # TYPICALITY (of row)
    # smoke tests
    order_by = False
    whereclause = None
    functions = bql.bql_statement.parseString('select name, qual_score, typicality from test',
                                              parseAll=True).functions
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    functions = bql.bql_statement.parseString('select name, qual_score, typicality from test',
                                              parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by typicality',
                                             parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    # TODO: test all other single-column functions
    # PROBABILITY <col>=<val>
    # PREDICTIVE PROBABILITY

    # TODO: test all single-column aggregate functions

    # TYPICALITY OF <col>
    functions = bql.bql_statement.parseString('select typicality of name from test',
                                              parseAll=True).functions
    order_by = bql.bql_statement.parseString('select * from test order by typicality',
                                             parseAll=True).order_by
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)

    # DEPENDENCE PROBABILITY OF <col> WITH <col> #DEPENDENCE PROBABILITY TO <col>
    # MUTUAL INFORMATION OF <col> WITH <col> #MUTUAL INFORMATION WITH <col>
    # CORRELATION OF <col> WITH <col>

    # TODO: test ordering by functions

@notimplemented
def test_delete_model():
    pass  # TODO


def test_update_schema():
    test_tablename, _ = create_dha()
    m_c, m_r, t = engine.persistence_layer.get_metadata_and_table(test_tablename)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['qual_score']] == 'numerical'
    assert cctypes[m_c['name_to_idx']['name']] == 'categorical'

    # Categorical with no cardinality parameter
    mappings = dict(qual_score=dict(cctype='categorical', parameters=None))
    engine.update_schema(test_tablename, mappings)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['qual_score']] == 'categorical'

    # Categorical with optional cardinality parameter
    mappings = dict(name=dict(cctype='categorical', parameters=dict(cardinality=350)))
    engine.update_schema(test_tablename, mappings)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['name']] == 'categorical'

    # Cyclic with required min, max parameters
    mappings = dict(qual_score=dict(cctype='cyclic', parameters=dict(min=0, max=100)))
    engine.update_schema(test_tablename, mappings)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['qual_score']] == 'cyclic'

    # Ignore
    mappings = dict(qual_score=dict(cctype='ignore', parameters=None))
    engine.update_schema(test_tablename, mappings)
    m_c, m_r, t = engine.persistence_layer.get_metadata_and_table(test_tablename)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert 'qual_score' not in m_c['name_to_idx'].keys()

    # Now test that it doesn't allow name to be continuous
    mappings = dict(name=dict(cctype='numerical', parameters=None))
    with pytest.raises(ValueError):
        engine.update_schema(test_tablename, mappings)

    # Test that setting cyclic with parameters inside the range of data fails
    mappings = dict(qual_score=dict(cctype='cyclic', parameters=dict(min=50, max=60)))
    with pytest.raises(utils.BayesDBError):
        engine.update_schema(test_tablename, mappings)

    # Test that setting categorical with low cardinality fails
    mappings = dict(name=dict(cctype='categorical', parameters=dict(cardinality=10)))
    with pytest.raises(utils.BayesDBError):
        engine.update_schema(test_tablename, mappings)


def test_save_and_load_models():
    test_tablename, _ = create_dha()
    engine.initialize_models(test_tablename, 3)
    engine.analyze(test_tablename, model_indices='all', iterations=1, background=False)
    # note that this won't save the models, since we didn't call this from the client.
    # engine.save_models actually just returns the models.
    original_models = engine.save_models(test_tablename)

    test_tablename2, _ = create_dha()
    models = original_models['models']
    model_schema = original_models['schema']
    engine.load_models(test_tablename2, models, model_schema)
    assert engine.save_models(test_tablename2).values() == original_models.values()


def test_initialize_models():
    test_tablename, _ = create_dha(path='data/dha_missing.csv')

    engine = Engine(seed=0)
    num_models = 5
    engine.initialize_models(test_tablename, num_models)

    model_ids = engine.persistence_layer.get_model_ids(test_tablename)
    assert sorted(model_ids) == range(num_models)
    for i in range(num_models):
        model = engine.persistence_layer.get_models(test_tablename, i)
        assert model['iterations'] == 0


def test_analyze():
    test_tablename, _ = create_dha()
    num_models = 3
    engine.initialize_models(test_tablename, num_models)

    for it in (1, 2):
        engine.analyze(test_tablename, model_indices='all', iterations=1, background=True)

        while 'not currently being analyzed' not in engine.show_analyze(test_tablename)['message']:
            import time
            time.sleep(0.1)

        # analyze_results = engine.show_analyze(test_tablename)
        model_ids = engine.persistence_layer.get_model_ids(test_tablename)
        assert sorted(model_ids) == range(num_models)
        for i in range(num_models):
            model = engine.persistence_layer.get_models(test_tablename, i)
            assert model['iterations'] == it

    for it in (3, 4):  # models were analyzed by previous for loop, so start counting at 3.
        engine.analyze(test_tablename, model_indices='all', iterations=1, background=False)
        analyze_results = engine.show_analyze(test_tablename)
        assert 'not currently being analyzed' in analyze_results['message']
        model_ids = engine.persistence_layer.get_model_ids(test_tablename)
        assert sorted(model_ids) == range(num_models)
        for i in range(num_models):
            model = engine.persistence_layer.get_models(test_tablename, i)
            assert model['iterations'] == it


def test_subsampling():
    # Use Kiva table, which has 10000 rows, instead of DHA.
    test_tablename = 'kivatest' + str(int(time.time() * 1000000)) + \
        str(int(random.random()*10000000))
    global test_tablenames
    test_tablenames.append(test_tablename)

    path = 'data/kiva_small.csv'
    header, rows = data_utils.read_csv(path)

    num_rows = 4  # rows in kiva_small
    num_rows_subsample = 2

    # client('create btable %s from %s' % (test_tablename, path), debug=True, pretty=False)
    # only analyze using some rows
    engine.create_btable(test_tablename, header, rows, subsample=num_rows_subsample, key_column=0)
    # make sure select (using no models) works and returns the correct number of rows
    functions = bql.bql_statement.parseString('select loan_id, loan_status from test',
                                              parseAll=True).functions
    whereclause = None
    limit = float('inf')
    order_by = False
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert len(select_result['data']) == num_rows  # number of rows in Kiva

    # TODO: better testing to see what we can do before subsampling (with partial models)

    num_models = 2
    iterations = 1
    engine.initialize_models(test_tablename, num_models)
    # analyze segfaults
    engine.analyze(test_tablename, model_indices='all', iterations=iterations, background=False)
    print('analyzed')
    model_ids = engine.persistence_layer.get_model_ids(test_tablename)
    for i in range(num_models):
        model = engine.persistence_layer.get_models(test_tablename, i)
        assert model['iterations'] == iterations

    # make sure normal queries work and return the correct number of rows
    functions = bql.bql_statement.parseString('select loan_id, predictive probability of '
                                              'loan_status from test', parseAll=True).functions
    whereclause = None
    limit = float('inf')
    order_by = False
    select_result = engine.select(test_tablename, functions, whereclause, limit, order_by, None)
    assert len(select_result['data']) == num_rows  # number of rows in Kiva


def test_nan_handling():
    test_tablename1, _ = create_dha(path='data/dha_missing.csv')
    test_tablename2, _ = create_dha(path='data/dha_missing_nan.csv')
    m1 = engine.persistence_layer.get_metadata(test_tablename1)
    m2 = engine.persistence_layer.get_metadata(test_tablename2)
    assert m1['M_c'] == m2['M_c']
    assert m1['M_r'] == m2['M_r']
    assert m1['cctypes'] == m2['cctypes']
    numpy.testing.assert_equal(numpy.array(m1['T']), numpy.array(m2['T']))


def test_infer():
    # TODO: whereclauses
    test_tablename, _ = create_dha(path='data/dha_missing.csv')

    # dha_missing has missing qual_score in first 5 rows, and missing name in rows 6-10.
    engine = Engine(seed=0)
    engine.initialize_models(test_tablename, 20)

    functions = bql.bql_statement.parseString('infer name, qual_score from test',
                                              parseAll=True).functions
    whereclause = None
    limit = float('inf')
    order_by = False
    numsamples = 30
    confidence = 0
    infer_result = engine.infer(test_tablename, functions, confidence, whereclause, limit,
                                numsamples, order_by)
    assert 'column_labels' in infer_result
    assert 'data' in infer_result
    assert infer_result['column_labels'] == ['key', 'name', 'qual_score']
    # 307 is the total number of rows in the dataset.
    assert(len(infer_result['data']) == 307 and
           len(infer_result['data'][0]) == len(infer_result['column_labels']))
    assert type(infer_result['data'][0][0]) == numpy.string_  # type of key is int
    t = type(infer_result['data'][0][1])
    assert (t == unicode) or (t == numpy.string_)  # type of name is string
    assert type(infer_result['data'][0][2]) == float  # type of qual_score is float

    all_possible_names = [infer_result['data'][row][1] for row in range(5) + range(10, 307)]
    all_observed_qual_scores = [infer_result['data'][row][2] for row in range(5, 307)]

    for row in range(5):
        inferred_name = infer_result['data'][row+5][1]
        inferred_qual_score = infer_result['data'][row][2]
        assert inferred_name in all_possible_names
        assert type(inferred_qual_score) == type(1.2)
        assert inferred_qual_score > min(all_observed_qual_scores)
        assert inferred_qual_score < max(all_observed_qual_scores)

    # Now, try infer with higher confidence, and make sure that name isn't inferred anymore.
    confidence = 0.9
    infer_result = engine.infer(test_tablename, functions, confidence, whereclause, limit,
                                numsamples, order_by)

    for row in range(5):
        # TODO: what do missing values look like? these should be missing
        inferred_name = infer_result['data'][row+5][1]
        inferred_qual_score = infer_result['data'][row][2]
        assert numpy.isnan(inferred_name)
        assert numpy.isnan(inferred_qual_score)


def test_simulate():
    # TODO: whereclauses
    test_tablename, _ = create_dha()
    engine.initialize_models(test_tablename, 2)

    columnstring = 'name, qual_score'
    functions = bql.bql_statement.parseString('simulate name, qual_score from test',
                                              parseAll=True).functions
    whereclause = None
    givens = None
    order_by = False
    numpredictions = 10
    simulate_result = engine.simulate(test_tablename, functions, givens, numpredictions, order_by)
    assert 'column_labels' in simulate_result
    assert 'data' in simulate_result
    assert simulate_result['column_labels'] == ['name', 'qual_score']

    assert(len(simulate_result['data']) == 10 and
           len(simulate_result['data'][0]) == len(simulate_result['column_labels']))
    for row in range(numpredictions):
        t = type(simulate_result['data'][row][0])
        assert (t == unicode) or (t == numpy.string_)
        assert type(simulate_result['data'][row][1]) == float


def test_estimate_pairwise_dependence_probability():
    test_tablename, _ = create_dha()
    engine.initialize_models(test_tablename, 2)
    dep_mat = engine.estimate_pairwise(test_tablename, 'dependence probability')


@pytest.mark.skipif(True, reason="Calculation too slow due to analysis on non-ignored, unique, "
                    "multinomial 'name' variable")
def test_estimate_pairwise_mutual_information():
    test_tablename, _ = create_dha()
    engine.initialize_models(test_tablename, 2)
    mi_mat = engine.estimate_pairwise(test_tablename, 'mutual information', numsamples=2)


def test_estimate_pairwise_correlation():
    test_tablename, _ = create_dha()
    engine.initialize_models(test_tablename, 2)
    cor_mat = engine.estimate_pairwise(test_tablename, 'correlation')


def test_list_btables():
    list_btables_result = engine.list_btables()['data']
    assert type(list_btables_result) == list
    initial_btable_count = len(list_btables_result)

    test_tablename1, create_btable_result = create_dha()
    test_tablename2, create_btable_result = create_dha()

    list_btables_result = engine.list_btables()['data']
    assert [test_tablename1] in list_btables_result
    assert [test_tablename2] in list_btables_result
    assert len(list_btables_result) == 2 + initial_btable_count

    engine.drop_btable(test_tablename1)
    test_tablename3, create_btable_result = create_dha()
    list_btables_result = engine.list_btables()['data']
    assert [test_tablename1] not in list_btables_result
    assert [test_tablename3] in list_btables_result
    assert [test_tablename2] in list_btables_result

    engine.drop_btable(test_tablename2)
    engine.drop_btable(test_tablename3)

    list_btables_result = engine.list_btables()['data']
    assert len(list_btables_result) == 0 + initial_btable_count


@notimplemented
def test_execute_file():
    pass  # TODO


def test_show_schema():
    test_tablename, _ = create_dha()
    m_c, m_r, t = engine.persistence_layer.get_metadata_and_table(test_tablename)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['qual_score']] == 'numerical'
    assert cctypes[m_c['name_to_idx']['name']] == 'categorical'

    schema = engine.show_schema(test_tablename)
    cctypes_full = engine.persistence_layer.get_cctypes_full(test_tablename)
    assert sorted([d[1] for d in schema['data']]) == sorted(cctypes_full)
    assert schema['data'][0][0] == 'key'

    mappings = dict(qual_score=dict(cctype='categorical', parameters=None))
    engine.update_schema(test_tablename, mappings)
    cctypes = engine.persistence_layer.get_cctypes(test_tablename)
    assert cctypes[m_c['name_to_idx']['qual_score']] == 'categorical'

    schema = engine.show_schema(test_tablename)
    cctypes_full = engine.persistence_layer.get_cctypes_full(test_tablename)
    assert sorted([d[1] for d in schema['data']]) == sorted(cctypes_full)
    assert schema['data'][0][0] == 'key'


def test_show_models():
    test_tablename, _ = create_dha()
    num_models = 3
    engine.initialize_models(test_tablename, num_models)

    for it in (1, 2):
        analyze_out = engine.analyze(test_tablename, model_indices='all', iterations=1,
                                     background=False)
        model_ids = engine.persistence_layer.get_model_ids(test_tablename)
        assert sorted(model_ids) == range(num_models)
        for i in range(num_models):
            model = engine.persistence_layer.get_models(test_tablename, i)
            assert model['iterations'] == it

        # models should be a list of (id, iterations) tuples.
        models = engine.show_models(test_tablename)['models']
        assert len(models) == num_models
        for iter_id, m in enumerate(models):
            assert iter_id == m[0]
            assert it == m[1]


def test_describe():
    # note that the input data does not have all the values that are in the value maps
    test_tablename, metadata = create_describe_btable()

    bql_string = 'describe c_0 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)

    assert(len(result['data']) == 1)
    assert(result['data'][0][0] == 'c_0')
    assert(result['data'][0][1] == 'column zero')
    assert(result['data'][0][2] == 'description for column zero')

    # check that M_c reflects the value maps
    metadata = engine.persistence_layer.get_metadata(test_tablename)
    col_0_idx = metadata['M_c']['name_to_idx']['c_0']
    assert(metadata['M_c']['column_metadata'][col_0_idx]['parameters']['cardinality'] == 6)
    assert(len(metadata['M_c']['column_metadata'][col_0_idx]['value_to_code']) == 6)
    assert(len(metadata['M_c']['column_metadata'][col_0_idx]['code_to_value']) == 6)

    assert('6' in result['data'][0][3])  # cardinality


    bql_string = 'describe c_1, c_2, c_3 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)

    assert(len(result['data']) == 3)
    for row in result['data']:
        assert(len(row) == 4)

    bql_string = 'describe * for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)

    assert(len(result['data']) == 4)
    for row in result['data']:
        assert(len(row) == 4)


def test_update_descriptions_single():
    test_tablename, metadata = create_describe_btable()

    description_proposed = 'Hamish the cat'

    bql_string = 'update description for %s set c_0="%s"' % (test_tablename, description_proposed)
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    _, args, _ = parser.parse_update_descriptions(bql_query)
    result = engine.update_descriptions(test_tablename, args['mappings'])

    column_name_output = result['data'][0][0]
    description_output = result['data'][0][1]

    assert(column_name_output == 'c_0')
    assert(description_output == description_proposed)

    bql_string = 'describe c_0 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)
    description_updated = result['data'][0][2]

    assert(description_updated == description_proposed)


def test_update_descriptions_multiple():
    test_tablename, metadata = create_describe_btable()

    description_proposed_0 = 'Hamish thh cat'
    description_proposed_1 = 'Winter Ninjaturtle'

    bql_string = 'update descriptions for %s set c_0="%s", c_2="%s"' \
        % (test_tablename, description_proposed_0, description_proposed_1)
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    _, args, _ = parser.parse_update_descriptions(bql_query)
    result = engine.update_descriptions(test_tablename, args['mappings'])

    column_name_output_0 = result['data'][0][0]
    description_output_0 = result['data'][0][1]
    column_name_output_1 = result['data'][1][0]
    description_output_1 = result['data'][1][1]

    assert(column_name_output_0 == 'c_0')
    assert(description_output_0 == description_proposed_0)
    assert(column_name_output_1 == 'c_2')
    assert(description_output_1 == description_proposed_1)

    bql_string = 'describe c_0, c_2 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)
    description_updated_0 = result['data'][0][2]
    description_updated_1 = result['data'][1][2]

    assert(description_updated_0 == description_proposed_0)
    assert(description_updated_1 == description_proposed_1)


def test_update_short_names_single():
    test_tablename, metadata = create_describe_btable()

    short_name_proposed = 'Hamish'

    bql_string = 'update short name for %s set c_0=%s' % (test_tablename, short_name_proposed)
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    _, args, _ = parser.parse_update_short_names(bql_query)
    result = engine.update_short_names(test_tablename, args['mappings'])

    column_name_output = result['data'][0][0]
    short_name_output = result['data'][0][1]

    assert(column_name_output == 'c_0')
    assert(short_name_output == short_name_proposed)

    bql_string = 'describe c_0 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)
    short_name_updated = result['data'][0][1]

    assert(short_name_updated == short_name_proposed)

@notimplemented
def test_update_codebook():
    pass # TODO

def test_update_short_names_multiple():
    test_tablename, metadata = create_describe_btable()

    short_name_proposed_0 = 'Hamish'
    short_name_proposed_1 = 'Winter Ninjaturtle'

    bql_string = 'update short names for %s set c_0="%s", c_2="%s"' \
        % (test_tablename, short_name_proposed_0, short_name_proposed_1)
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    _, args, _ = parser.parse_update_short_names(bql_query)
    result = engine.update_short_names(test_tablename, args['mappings'])

    column_name_output_0 = result['data'][0][0]
    short_name_output_0 = result['data'][0][1]
    column_name_output_1 = result['data'][1][0]
    short_name_output_1 = result['data'][1][1]

    assert(column_name_output_0 == 'c_0')
    assert(short_name_output_0 == short_name_proposed_0)
    assert(column_name_output_1 == 'c_2')
    assert(short_name_output_1 == short_name_proposed_1)

    bql_string = 'describe c_0, c_2 for %s' % test_tablename
    bql_query = bql.bql_statement.parseString(bql_string, parseAll=True)
    result = engine.describe(test_tablename, bql_query.columnset)
    short_name_updated_0 = result['data'][0][1]
    short_name_updated_1 = result['data'][1][1]

    assert(short_name_updated_0 == short_name_proposed_0)
    assert(short_name_updated_1 == short_name_proposed_1)


def test_show_diagnostics():
    test_tablename, _ = create_describe_btable()

    with pytest.raises(utils.BayesDBError) as excinfo:
        result = engine.show_diagnostics(test_tablename)
    assert 'No models for btable' in excinfo.value.message

    engine.initialize_models(test_tablename, n_models=2)

    with pytest.raises(utils.BayesDBError) as excinfo:
        result = engine.show_diagnostics(test_tablename)
    assert 'No diagnostics found' in excinfo.value.message

    num_iters = 5
    engine.analyze(test_tablename, iterations=num_iters, background=False)

    results = engine.show_diagnostics(test_tablename)

    # there should be an entry for each diagnostic and also: model id, iterations, time, and
    # logscore
    entry_length = len(diag_utils.single_state_diagnostics) + 4
    assert(len(results['column_labels']) == entry_length)
    assert(len(results['data']) == 2)
    assert(len(results['data'][0]) == entry_length)
    assert(len(results['data'][1]) == entry_length)

    # check values for id and iterations
    assert(results['data'][0][0] == 0)
    assert(results['data'][0][1] == num_iters)

    assert(results['data'][1][0] == 1)
    assert(results['data'][1][1] == num_iters)


@notimplemented
def test_drop_models():
    pass  # TODO


def test_estimate_columns():
    # TODO: add nontrivial cases
    test_tablename, _ = create_dha()
    metadata = engine.persistence_layer.get_metadata(test_tablename)
    all_columns = metadata['M_c']['name_to_idx'].keys()
    engine.initialize_models(test_tablename, 2)

    whereclause = None
    limit = float('inf')
    order_by = False
    name = None
    functions = None
    column_labels = engine.estimate_columns(test_tablename, functions, whereclause, limit, order_by,
                                            name)['column_labels']
    assert column_labels == ['column label', 'column name']

if __name__ == '__main__':
    run_test()
